import numpy as np
import pytest

import dclab
from dclab import new_dataset
from dclab.features import inert_ratio as ir

from helper_methods import retrieve_data


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'feat_anc_core.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_af_inert_ratio_cvx():
    pytest.importorskip("nptdms")
    # Brightness of the image
    ds = dclab.new_dataset(retrieve_data("fmt-tdms_fl-image-bright_2017.zip"))
    # This is something low-level and should not be done in a script.
    # Remove the brightness columns from RTDCBase to force computation with
    # the image and contour columns.
    real_ir = ds._events.pop("inert_ratio_cvx")
    # This will cause a zero-padding warning:
    comp_ir = ds["inert_ratio_cvx"]
    idcompare = ~np.isnan(comp_ir)
    # ignore first event (no image data)
    idcompare[0] = False
    assert np.allclose(real_ir[idcompare], comp_ir[idcompare])


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'feat_anc_core.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_af_inert_ratio_prnc():
    pytest.importorskip("nptdms")
    # Brightness of the image
    ds = dclab.new_dataset(retrieve_data("fmt-tdms_fl-image-bright_2017.zip"))
    # This will cause a zero-padding warning:
    prnc = ds["inert_ratio_prnc"]
    raw = ds["inert_ratio_raw"]
    idcompare = ~np.isnan(prnc)
    # ignore first event (no image data)
    idcompare[0] = False
    diff = (prnc - raw)[idcompare]
    # only compare the first valid event which seems to be quite close
    assert np.allclose(diff[0], 0, atol=1.2e-3, rtol=0)


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'feat_anc_core.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_af_inert_ratio_raw():
    pytest.importorskip("nptdms")
    # Brightness of the image
    ds = dclab.new_dataset(retrieve_data("fmt-tdms_fl-image-bright_2017.zip"))
    # This is something low-level and should not be done in a script.
    # Remove the brightness columns from RTDCBase to force computation with
    # the image and contour columns.
    real_ir = ds._events.pop("inert_ratio_raw")
    # This will cause a zero-padding warning:
    comp_ir = ds["inert_ratio_raw"]
    idcompare = ~np.isnan(comp_ir)
    # ignore first event (no image data)
    idcompare[0] = False
    assert np.allclose(real_ir[idcompare], comp_ir[idcompare])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_inert_ratio_raw():
    ds = new_dataset(retrieve_data("fmt-hdf5_mask-contour_2018.zip"))

    raw = ir.get_inert_ratio_raw(cont=ds["contour"])
    ref = np.array([4.25854232,  1.22342663,  4.64971179,  1.70914857,
                    3.62797492, 1.51502192,  2.74757573,  1.79841136])
    assert np.allclose(ref, raw, rtol=0, atol=5e-9)


def test_inert_ratio_prnc():
    """Test equivalence of inert_ratio_raw and inert_ratio_prnc"""
    t = np.linspace(0, 2*np.pi, 300)

    x1 = 1.7 * np.cos(t)
    y1 = 1.1 * np.sin(t)
    c1 = np.dstack((x1, y1))[0]

    phi = np.arctan2(y1, x1)
    rho = np.sqrt(x1**2 + y1**2)

    for theta in np.linspace(0.1, 2*np.pi, 14):  # arbitrary rotation
        for pos_x in np.linspace(-5, 20, 8):  # arbitrary x shift
            for pos_y in np.linspace(-4.6, 17, 4):  # arbitrary y shift
                x2 = rho * np.cos(phi + theta) + pos_x
                y2 = rho * np.sin(phi + theta) + pos_y

                c2 = np.dstack((x2, y2))[0]
                raw = ir.get_inert_ratio_raw(c1)
                prnc = ir.get_inert_ratio_prnc(c2)

                assert np.allclose(raw, prnc, rtol=0, atol=1e-10)


def test_inert_ratio_prnc_simple_1():
    c = np.array([[0, 0],
                  [0, 1],
                  [0, 2],
                  [1, 2],
                  [2, 2],
                  [3, 2],
                  [3, 1],
                  [3, 0],
                  [2, 0],
                  [1, 0],
                  [0, 0]])
    raw = ir.get_inert_ratio_raw(c)
    prnc = ir.get_inert_ratio_prnc(c)
    tilt = ir.get_tilt(c)
    assert np.allclose(raw, 1.5)
    assert np.allclose(prnc, 1.5)
    assert np.allclose(tilt, 0)


def test_inert_ratio_prnc_simple_2():
    c = np.array([[0, 0],
                  [1, 1],
                  [2, 2],
                  [3, 3],
                  [4, 2],
                  [5, 1],
                  [4, 0],
                  [3, -1],
                  [2, -2],
                  [1, -1],
                  [0, 0]])
    raw = ir.get_inert_ratio_raw(c)
    prnc = ir.get_inert_ratio_prnc(c)
    tilt = ir.get_tilt(c)
    assert np.allclose(raw, 1)
    assert np.allclose(prnc, 1.5)
    assert np.allclose(tilt, np.pi/4)


def test_tilt():
    t = np.linspace(0, 2*np.pi, 300)

    x1 = 1.7 * np.cos(t)
    y1 = 1.1 * np.sin(t)

    phi = np.arctan2(y1, x1)
    rho = np.sqrt(x1**2 + y1**2)

    for theta in np.linspace(-.3, 2.2*np.pi, 32):  # arbitrary rotation
        x2 = rho * np.cos(phi + theta)
        y2 = rho * np.sin(phi + theta)

        c2 = np.dstack((x2, y2))[0]
        tilt = ir.get_tilt(c2)

        th = np.mod(theta, np.pi)
        if th > np.pi/2:
            th = np.pi - th
        assert np.allclose(tilt, th)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Overview plot for the Young's modulus look-up table

The Young's modulus is corrected for pixelation (pixel size 0.34 µm).
The temperature is set to 23°C.
The 20µm plot is used in the dclab documentation. The others
are used in the Shape-Out documentation.
"""
from dclab.features import emodulus
import numpy as np
import seaborn as sns
import matplotlib.pylab as plt
from matplotlib import cm, colors
from matplotlib.ticker import MultipleLocator

sns.set_style("whitegrid")


lut_name = "LE-2D-FEM-19"
media = ["CellCarrier", "CellCarrier B", "water"]
matrix = {  # channel width: flow rates
    15: [0.016, 0.032, 0.048],
    20: [0.040, 0.080, 0.120],
    30: [0.160, 0.240, 0.320],
    40: [0.320, 0.400, 0.600],
}
deform_max = 0.17
grid_size = 1000
emin = 4e-2
emax = 2e1
norm = colors.LogNorm(vmin=emin, vmax=emax)
cmap = "jet"

for channel_width in matrix:
    flow_rates = matrix[channel_width]
    # setup figure
    fig, axes = plt.subplots(nrows=3, ncols=3, sharex=True, sharey=True,
                             gridspec_kw={"left": 0.12,
                                          "right": .85,
                                          "wspace": .1,
                                          "hspace": .15,
                                          },
                             figsize=(7.5, 6.3))
    fig.suptitle("{} µm channel".format(channel_width))
    axes[2][1].set_xlabel("Area [µm²]")
    axes[1][0].set_ylabel("Deformation")
    # column label (flow rate)
    for ii in range(3):
        axes[0][ii].set_title("{} µL/s".format(flow_rates[ii]),
                              fontsize=10, fontweight="bold")
    # row label (medium)
    for jj in range(3):
        pos = axes[jj][0].get_position()
        x = pos.x0 - .11
        y = pos.y0 + pos.height/2
        fig.text(x, y, media[jj], ha='left', va='center', rotation=90,
                 fontsize=10, fontweight="bold")

    # plot the LUTs channel-wise
    lut_base, lut_meta = emodulus.load_lut(lut_name)
    deform = np.linspace(lut_base[:, 1].min(), lut_base[:, 1].max(),
                         grid_size, endpoint=True)
    area_um = np.linspace(lut_base[:, 0].min(), lut_base[:, 0].max(),
                          grid_size, endpoint=True)
    for jj, medium in enumerate(media):  # rows
        for ii, flow_rate in enumerate(matrix[channel_width]):  # columns
            area_um_scaled = emodulus.scale_linear.scale_area_um(
                area_um=area_um,
                channel_width_in=lut_meta["channel_width"],
                channel_width_out=channel_width)
            aa, dd = np.meshgrid(area_um_scaled, deform)
            emod = emodulus.get_emodulus(area_um=aa,
                                         deform=dd,
                                         medium=medium,
                                         channel_width=channel_width,
                                         flow_rate=flow_rate,
                                         px_um=.34,
                                         temperature=23.0,
                                         lut_data=lut_name)
            ax = axes[jj][ii]
            ax.set_aspect("auto")
            pc = ax.pcolormesh(aa, dd, emod, shading="nearest", cmap=cmap,
                               norm=norm, zorder=11)
            ax.grid(zorder=10)
            ax.yaxis.set_major_locator(MultipleLocator(0.05))  # ticks every .5

    axes[2][2].set_ylim(0, axes[2][2].get_ylim()[1] + .03)
    xmin, xmax = axes[2][2].get_xlim()
    axes[2][2].set_xlim(xmin - (xmax-xmin)/30, xmax + (xmax-xmin)/30)
    axes[2][2].grid(True, which="minor", zorder=10)

    # add colorbar
    cax = plt.axes([0.87, 0.1, 0.02, 0.8])

    mappable = cm.ScalarMappable(norm=norm, cmap=cmap)
    plt.colorbar(pc, orientation="vertical", cax=cax, extend="both",
                 label="apparent Young's modulus ({}) [kPa]".format(lut_name))

    plt.savefig("_emodulus_{}um_{}.png".format(channel_width, lut_name),
                dpi=150)
    plt.close()



"""Check RT-DC datasets for completeness"""
import copy
import functools
import warnings

import h5py
import numpy as np

from .core import RTDCBase
from .fmt_hierarchy import RTDC_Hierarchy
from .load import load_file

from .. import definitions as dfn

#: These sections should be fully present, except for the
#: keys in :data:`OPTIONAL_KEYS`.
DESIRABLE_SECTIONS = {
    "experiment",
    "imaging",
    "online_contour",
    "setup",
}

#: log names that end with these strings are not checked
IGNORED_LOG_NAMES = {
    "_para.ini",
    "_image.ini",
    "FG_Config.mcf",
    "parameters.txt",
    "_SoftwareSettings.ini",
    "dckit-history",
}

#: keys that must be present for every measurement
IMPORTANT_KEYS = {
    "experiment": [
        "date",
        "event count",
        "run index",
        "sample",
        "time"],
    "imaging": [
        "flash device",
        "flash duration",
        "frame rate",
        "pixel size",
        "roi position x",
        "roi position y",
        "roi size x",
        "roi size y"],
    "setup": [
        "channel width",
        "chip region",
        "flow rate",
        "medium"],
}

#: keys that must be present for fluorescence measurements
IMPORTANT_KEYS_FL = {
    "fluorescence": [
        "bit depth",
        "channel count",
        "channels installed",
        "laser count",
        "lasers installed",
        "sample rate",
        "samples per event",
        "signal max",
        "signal min",
        "trace median"],
}

#: maximum line length in log files
LOG_MAX_LINE_LENGTH = 100

#: keys that are optional
OPTIONAL_KEYS = {
    "fluorescence": [
        "baseline 1 offset",
        "baseline 2 offset",
        "baseline 3 offset",
        # name, lambda, power have their own special tests
        "channel 1 name",
        "channel 2 name",
        "channel 3 name",
        "laser 1 lambda",
        "laser 2 lambda",
        "laser 3 lambda",
        "laser 1 power",
        "laser 2 power",
        "laser 3 power",
    ],
    "setup": [
        "temperature",
        "chip identifier",
    ],
    "online_contour": [
        # introduced in 0.34.0
        "bg empty",
    ],
}

#: valid metadata choices
#: .. versionchanged:: 0.29.1
#:    medium not anymore restricted to certain set of choices
VALID_CHOICES = {}


@functools.total_ordering
class ICue(object):
    def __init__(self, msg, level, category, data=None, identifier=None,
                 cfg_section=None, cfg_key=None, cfg_choices=None):
        """Integrity cue"""
        #: human readable message
        self.msg = msg
        #: severity level ("violation", "alert", or "info")
        self.level = level
        #: machine-readable data associated with the check
        self.data = data
        #: fail category
        self.category = category
        #: identifier e.g. for UI manipulation in DCKit
        self.identifier = identifier
        #: section (only for categories "missing metadata", "bad metadata")
        self.cfg_section = cfg_section
        #: key (only for categories "missing metadata", "bad metadata");
        #: can be omitted e.g. to communicate that the entire section is
        #: missing
        self.cfg_key = cfg_key
        #: allowed choices for the specific [section]: key combination
        #: (only for categories "missing metadata", "bad metadata")
        self.cfg_choices = cfg_choices
        if self.cfg_choices is None:
            if (cfg_section in VALID_CHOICES
                    and cfg_key in VALID_CHOICES[cfg_section]):
                self.cfg_choices = VALID_CHOICES[cfg_section][cfg_key]

    def __eq__(self, other):
        leveld = {"info": 0,
                  "violation": 1,
                  "alert": 2,
                  }
        return ((leveld[self.level], self.cfg_section or "",
                 self.cfg_key or "", self.category, self.msg) ==
                (leveld[other.level], other.cfg_section or "",
                 other.cfg_key or "", other.category, other.msg))

    def __lt__(self, other):
        leveld = {"info": 0,
                  "violation": 1,
                  "alert": 2, }
        return ((leveld[self.level], self.cfg_section or "",
                 self.cfg_key or "", self.category, self.msg) <
                (leveld[other.level], other.cfg_section or "",
                 other.cfg_key or "", other.category, other.msg))

    def __repr__(self):
        return "<ICue: '{}' at 0x{}>".format(self.msg, hex(id(self)))

    @staticmethod
    def get_level_summary(cues):
        """For a list of ICue, return the abundance of all levels"""
        levels = {"info": 0,
                  "alert": 0,
                  "violation": 0}
        for cue in cues:
            levels[cue.level] += 1
        return levels


class IntegrityChecker(object):
    def __init__(self, path_or_ds):
        """Check the integrity of a dataset

        The argument must be either a path to an .rtdc or .tdms file
        or an instance of `RTDCBase`. If a path is given, then all
        warnings (e.g. UnknownConfigurationKeyWarning) are catched
        and added to the cue list.

        Usage:

        .. code:: python

            ic = IntegrityChecker("/path/to/data.rtdc")
            cues = ic.check()

        """
        self.warn_cues = []
        if isinstance(path_or_ds, RTDCBase):
            self.ds = path_or_ds
            self.finally_close = False
        else:
            with warnings.catch_warnings(record=True) as ws:
                warnings.simplefilter("always")
                self.ds = load_file(path_or_ds)
                for ww in ws:
                    self.warn_cues.append(ICue(
                        msg="{}: {}".format(ww.category.__name__, ww.message),
                        level="alert",
                        category="warning"))
            self.finally_close = True

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        # close the file
        if self.finally_close and hasattr(self.ds, "__exit__"):
            self.ds.__exit__(type, value, tb)

    @property
    def has_fluorescence(self):
        if ("fluorescence" in self.ds
                or "fl1_max" in self.ds
                or "fl2_max" in self.ds
                or "fl3_max" in self.ds):
            fl = True
        else:
            fl = False
        return fl

    def check(self, **kwargs):
        """Run all checks

        This method calls all class methods that start with `check_`.
        `kwargs` are passed to all methods. Possible options are:

        - "expand_section" `bool`: add a cue for every missing
          metadata key if a section is missing
        """
        cues = []
        funcs = IntegrityChecker.__dict__
        if not len(self.ds) == np.sum(self.ds.filter.all):
            raise NotImplementedError(
                "Integrity checks for datasets with active event filters "
                "are not supported!")
        elif self.ds.__class__ == RTDC_Hierarchy:
            raise NotImplementedError(
                "Integrity checks for 'RTDC_Hierarchy' instances are "
                "not supported!")
        for ff in sorted(funcs.keys()):
            if ff.startswith("check_fl_") and not self.has_fluorescence:
                # skip
                continue
            elif ff.startswith("check_"):
                cues += funcs[ff](self, **kwargs)
        return sorted(self.warn_cues + cues)

    def check_compression(self, **kwargs):
        cues = []
        if self.ds.format == "tdms":
            compression = "None"
            data = {"compressed": 0,
                    "total": 1,
                    "uncompressed": 0,
                    }
        elif self.ds.format == "hdf5":
            def iter_count_compression(h5):
                comp = 0
                noco = 0
                for key in h5:
                    obj = h5[key]
                    if isinstance(obj, h5py.Dataset):
                        if obj.compression is None:
                            noco += 1
                        else:
                            comp += 1
                    elif isinstance(obj, h5py.Group):
                        coi, noi = iter_count_compression(obj)
                        comp += coi
                        noco += noi
                    else:
                        raise ValueError("Unknown object: {}".format(obj))
                return comp, noco

            comp, noco = iter_count_compression(self.ds._h5)
            if noco == 0:
                compression = "All"
            elif comp == 0:
                compression = "None"
            else:
                compression = "Partial ({} of {})".format(comp, noco + comp)
            data = {"compressed": comp,
                    "total": noco + comp,
                    "uncompressed": noco,
                    }
        else:
            compression = "Unknown"
            data = None
        cues.append(ICue(
            msg="Compression: {}".format(compression),
            level="info",
            category="general",
            data=data))
        return cues

    def check_feat_index(self, **kwargs):
        """Up until"""
        cues = []
        lends = len(self.ds)
        if "index" in self.ds:
            if not np.all(self.ds["index"] == np.arange(1, lends + 1)):
                cues.append(ICue(
                    msg="The index feature is not enumerated correctly",
                    level="violation",
                    category="feature data"))
        return cues

    def check_feature_size(self, **kwargs):
        cues = []
        lends = len(self.ds)
        for feat in self.ds.features_innate:
            msg = "Features: wrong event count: '{}' ({} of {})"
            if feat == "trace":
                for tr in list(self.ds["trace"].keys()):
                    if len(self.ds["trace"][tr]) != lends:
                        cues.append(ICue(
                            msg=msg.format("trace/" + tr,
                                           len(self.ds["trace"][tr]),
                                           lends),
                            level="violation",
                            category="feature size"))
            else:
                if len(self.ds[feat]) != lends:
                    cues.append(ICue(
                        msg=msg.format(feat, len(self.ds[feat]), lends),
                        level="violation",
                        category="feature size"))
        return cues

    def check_features_unknown_hdf5(self, **kwargs):
        """check for feature column names"""
        cues = []
        if self.ds.format == "hdf5":
            for feat in self.ds._h5["events"]:
                if not dfn.feature_exists(feat):
                    cues.append(ICue(
                        msg="Features: Unknown key '{}'".format(feat),
                        level="violation",
                        category="feature unknown"))
        return cues

    def check_fl_metadata_channel_names(self, **kwargs):
        cues = []
        for ii in range(1, 4):
            chn = "channel {} name".format(ii)
            fli = "fl{}_max".format(ii)
            if (fli in self.ds
                    and chn not in self.ds.config["fluorescence"]):
                # Channel names must be defined when there is
                # a corresponding fluorescence signal.
                cues.append(ICue(
                    msg="Metadata: Missing key [{}] '{}'".format(
                        "fluorescence", chn),
                    level="alert",
                    category="metadata missing",
                    cfg_section="fluorescence",
                    cfg_key=chn))
            elif (fli not in self.ds
                  and chn in self.ds.config["fluorescence"]):
                # Channel names must not be defined when there is
                # no corresponding fluorescence signal.
                cues.append(ICue(
                    msg="Metadata: Unused key defined [{}] '{}'".format(
                        "fluorescence", chn),
                    level="alert",
                    category="metadata invalid",
                    cfg_section="fluorescence",
                    cfg_key=chn))
        return cues

    def check_fl_num_channels(self, **kwargs):
        cues = []
        # check for number of channels
        if "channel count" in self.ds.config["fluorescence"]:
            chc1 = self.ds.config["fluorescence"]["channel count"]
            chc2 = 0
            for ii in range(1, 4):
                chn = "channel {} name".format(ii)
                ecn = "fl{}_max".format(ii)
                if (chn in self.ds.config["fluorescence"] and
                        ecn in self.ds._events):
                    chc2 += 1
            if chc1 != chc2:
                cues.append(ICue(
                    msg="Metadata: fluorescence channel count inconsistent",
                    level="violation",
                    category="metadata wrong",
                    cfg_section="fluorescence",
                    cfg_key="channel count"))
        return cues

    def check_fl_num_lasers(self, **kwargs):
        cues = []
        # check for number of lasers
        if "laser count" in self.ds.config["fluorescence"]:
            lsc1 = self.ds.config["fluorescence"]["laser count"]
            lsc2 = 0
            for ii in range(1, 4):
                kl = "laser {} lambda".format(ii)
                kp = "laser {} power".format(ii)
                if (kl in self.ds.config["fluorescence"] and
                        kp in self.ds.config["fluorescence"] and
                        self.ds.config["fluorescence"][kp] != 0):
                    lsc2 += 1
            if lsc1 != lsc2:
                cues.append(ICue(
                    msg="Metadata: fluorescence laser count inconsistent",
                    level="violation",
                    category="metadata wrong",
                    cfg_section="fluorescence",
                    cfg_key="laser count"))
        return cues

    def check_fl_samples_per_event(self, **kwargs):
        cues = []
        # check for samples per event
        if "samples per event" in self.ds.config["fluorescence"]:
            spe = self.ds.config["fluorescence"]["samples per event"]
            if "trace" in self.ds:
                for key in self.ds["trace"].keys():
                    spek = self.ds["trace"][key][0].size
                    if spek != spe:
                        cues.append(ICue(
                            msg="Metadata: wrong number of samples per "
                                + "event: {} (expected {}, got {})".format(
                                    key, spe, spek),
                            level="violation",
                            category="metadata wrong",
                            cfg_section="fluorescence",
                            cfg_key="samples per event"))
        return cues

    def check_fl_max_positive(self, **kwargs):
        """Check if all fl?_max values are >0.1"""
        cues = []
        neg_feats = []
        for fl in ['fl1_max', 'fl2_max', 'fl3_max']:
            if fl in self.ds:
                if min(self.ds[fl]) <= 0.1:
                    neg_feats.append(fl)
        if neg_feats:
            cues.append(ICue(
                msg="Negative value for feature(s): {}".format(
                    ", ".join(neg_feats)),
                level="alert",
                category="feature data"))
        return cues

    def check_fl_max_ctc_positive(self, **kwargs):
        """Check if all fl?_max_ctc values are > 0.1"""
        cues = []
        neg_feats = []
        for fl in ['fl1_max_ctc', 'fl2_max_ctc', 'fl3_max_ctc']:
            if fl in self.ds:
                if min(self.ds[fl]) <= 0.1:
                    neg_feats.append(fl)
        if neg_feats:
            cues.append(ICue(
                msg="Negative value for feature(s): {}".format(
                    ", ".join(neg_feats)),
                level="alert",
                category="feature data"))
        return cues

    def check_flow_rate(self, **kwargs):
        """Make sure sheath and sample flow rates add up"""
        cues = []
        if ("setup" in self.ds.config
                and "flow rate" in self.ds.config["setup"]
                and "flow rate sample" in self.ds.config["setup"]
                and "flow rate sheath" in self.ds.config["setup"]):
            frsum = self.ds.config["setup"]["flow rate"]
            frsam = self.ds.config["setup"]["flow rate sample"]
            frshe = self.ds.config["setup"]["flow rate sheath"]
            if not np.allclose(frsum, frsam + frshe):
                for k in ["flow rate", "flow rate sheath", "flow rate sample"]:
                    cues.append(ICue(
                        msg="Metadata: Flow rates don't add up (sheath "
                            + "{:g} + sample {:g} != channel {:g})".format(
                                frshe, frsam, frsum),
                        level="alert",
                        category="metadata wrong",
                        cfg_section="setup",
                        cfg_key=k))
        return cues

    def check_fmt_hdf5(self, **kwargs):
        cues = []
        # hdf5-based checks
        if self.ds.format == "hdf5":
            # check meta data of images
            for feat in ["image", "image_bg", "mask"]:
                if feat in self.ds._events:
                    imdat = self.ds[feat]
                    for key, val in [['CLASS', b'IMAGE'],
                                     ['IMAGE_VERSION', b'1.2'],
                                     ['IMAGE_SUBCLASS', b'IMAGE_GRAYSCALE']]:
                        if key not in imdat.attrs:
                            cues.append(ICue(
                                msg="HDF5: '/{}': missing ".format(feat)
                                    + "attribute '{}'".format(key),
                                level="alert",
                                category="format HDF5"))
                        elif imdat.attrs.get_id(key).dtype.char != "S":
                            cues.append(ICue(
                                msg="HDF5: '/{}': attribute '{}' ".format(
                                    feat, key)
                                + "should be fixed-length ASCII string",
                                level="alert",
                                category="format HDF5"))
                        elif imdat.attrs[key] != val:
                            cues.append(ICue(
                                msg="HDF5: '/{}': attribute '{}' ".format(
                                    feat, key)
                                + "should have value '{}'".format(val),
                                level="alert",
                                category="format HDF5"))
            # check length of logs
            with h5py.File(self.ds.path, mode="r") as h5:
                if "logs" in h5:
                    logs = h5["logs"]
                    for logname in logs.keys():
                        # ignore tmds meta data log files
                        lign = [logname.endswith(n) for n in IGNORED_LOG_NAMES]
                        if sum(lign):
                            continue
                        log = logs[logname]
                        for ii in range(len(log)):
                            if len(log[ii]) > LOG_MAX_LINE_LENGTH:
                                cues.append(ICue(
                                    msg="Logs: {} line {} ".format(logname, ii)
                                        + "exceeds maximum line length "
                                        + "{}".format(LOG_MAX_LINE_LENGTH),
                                    level="alert",
                                    category="format HDF5"))
        return cues

    def check_info(self, **kwargs):
        cues = []
        cues.append(ICue(
            msg="Fluorescence: {}".format(self.has_fluorescence),
            level="info",
            category="general"))
        cues.append(ICue(
            msg="Data file format: {}".format(self.ds.format),
            level="info",
            category="general"))
        return cues

    def check_metadata_bad(self, **kwargs):
        cues = []
        # check for ROI size
        if ("imaging" in self.ds.config
                and "roi size x" in self.ds.config["imaging"]
                and "roi size y" in self.ds.config["imaging"]):
            for ii, roi in enumerate(["roi size y", "roi size x"]):
                for feat in ["image", "image_bg", "mask"]:
                    if feat in self.ds:
                        soll = self.ds[feat].shape[ii+1]
                        ist = self.ds.config["imaging"][roi]
                        if soll != ist:
                            cues.append(ICue(
                                msg="Metadata: Mismatch [imaging] "
                                    + "'{}' and feature {} ".format(roi,
                                                                    feat)
                                    + "({} vs {})".format(ist, soll),
                                level="violation",
                                category="metadata wrong",
                                cfg_section="imaging",
                                cfg_key=roi))
        return cues

    def check_metadata_bad_greater_zero(self, **kwargs):
        cues = []
        # check for ROI size
        for sec, key in [
            ["imaging", "frame rate"],
            ["imaging", "pixel size"],
            ["setup", "channel width"],
            ["setup", "flow rate"],
        ]:
            value = self.ds.config.get(sec, {}).get(key)
            if value is not None and value <= 0:
                cues.append(ICue(
                    msg=f"Metadata: Invalid value for [{sec}] '{key}': "
                        + f"'{value}'!",
                    level="violation",
                    category="metadata wrong",
                    cfg_section=sec,
                    cfg_key=key))
        return cues

    def check_metadata_choices(self, **kwargs):
        cues = []
        for sec in VALID_CHOICES:
            for key in VALID_CHOICES[sec]:
                if sec in self.ds.config and key in self.ds.config[sec]:
                    val = self.ds.config[sec][key]
                    if val not in VALID_CHOICES[sec][key]:
                        cues.append(ICue(
                            msg="Metadata: Invalid value [{}] {}: '{}'".format(
                                sec, key, val),
                            level="violation",
                            category="metadata wrong",
                            cfg_section=sec,
                            cfg_key=key,
                            cfg_choices=VALID_CHOICES[sec][key]))
        return cues

    def check_metadata_hdf5_type(self, **kwargs):
        """This is a low-level HDF5 check"""
        cues = []
        if self.ds.format == "hdf5":
            for entry in self.ds._h5.attrs:
                if entry.count(":"):
                    sec, key = entry.split(":")
                    val_act = self.ds._h5.attrs[entry]  # actual value
                    if isinstance(val_act, bytes):
                        val_act = val_act.decode("utf-8")
                    # Check whether the config key exists
                    if dfn.config_key_exists(sec, key):
                        func = dfn.get_config_value_func(sec, key)
                        val_exp = func(val_act)  # expected value
                        if (isinstance(val_exp, (list, np.ndarray))
                                and np.allclose(val_exp, val_act)):
                            continue
                        elif val_exp == val_act:
                            continue
                        else:
                            cues.append(ICue(
                                msg=f"Metadata: [{sec}]: '{key}' should be "
                                    + f"'{val_exp}', but is '{val_act}'!",
                                level="alert",
                                category="metadata wrong",
                                cfg_section=sec,
                                cfg_key=key))
        return cues

    def check_metadata_online_filter_polygon_points_shape(self, **kwargs):
        cues = []
        if "online_filter" in self.ds.config:
            for key in self.ds.config["online_filter"].keys():
                if key.endswith("polygon points"):
                    points = self.ds.config["online_filter"][key]
                    if points.shape[1] != 2 or points.shape[0] < 3:
                        cues.append(ICue(
                            msg="Metadata: Wrong shape [online_filter] "
                                + f"{key}: '{points.shape}'",
                            level="violation",
                            category="metadata wrong",
                            cfg_section="online_filter",
                            cfg_key=key))
        return cues

    def check_metadata_missing(self, expand_section=True, **kwargs):
        cues = []
        # These "must" be present:
        important = copy.deepcopy(IMPORTANT_KEYS)
        if self.has_fluorescence:
            important.update(IMPORTANT_KEYS_FL)
        # A list of sections we would like to investigate
        secs_investiage = list(set(important.keys()) | set(DESIRABLE_SECTIONS))

        for sec in secs_investiage:
            if sec not in self.ds.config and not expand_section:
                cues.append(ICue(
                    msg="Metadata: Missing section '{}'".format(sec),
                    level="violation" if sec in important else "alert",
                    category="metadata missing",
                    cfg_section=sec))
            else:
                for key in dfn.config_keys[sec]:
                    if key not in self.ds.config[sec]:
                        if sec in OPTIONAL_KEYS and key in OPTIONAL_KEYS[sec]:
                            # ignore this key
                            continue
                        elif sec in important and key in important[sec]:
                            level = "violation"
                        else:
                            level = "alert"
                        cues.append(ICue(
                            msg="Metadata: Missing key [{}] '{}'".format(sec,
                                                                         key),
                            level=level,
                            category="metadata missing",
                            cfg_section=sec,
                            cfg_key=key))
        # check for temperature
        if "temp" in self.ds:
            if "temperature" not in self.ds.config["setup"]:
                cues.append(ICue(
                    msg="Metadata: Missing key [setup] 'temperature', "
                        + "because the 'temp' feature is given",
                    level="alert",
                    category="metadata missing",
                    cfg_section="setup",
                    cfg_key="temperature"))
        return cues

    def check_ml_class(self, **kwargs):
        """Try to comput ml_class feature and display error message"""
        cues = []
        if "ml_class" in self.ds:
            try:
                self.ds["ml_class"]
            except ValueError as e:
                cues.append(ICue(
                    msg=e.args[0],
                    level="violation",
                    category="feature data"))
        return cues

    def check_shapein_issue3_bad_medium(self, **kwargs):
        """Some versions of Shape-In stored wrong [setup]: medium

        The problem only affects selection of "CellCarrier" which had
        index 0 and as a result "CellCarrierB" was written to the file.
        This means we only have to check for the Shape-In version and
        whether the medium is 'CellCarrierB'. In DCKit, the user can
        then manually edit the medium.

        Affected Shape-In versions: >=2.2.1.0,<2.2.2.3
        Also affected if bad config file was used: 2.2.2.4, 2.3.0.0

        https://github.com/ZELLMECHANIK-DRESDEN/ShapeIn_Issues/issues/3
        """
        cues = []
        medium = self.ds.config["setup"].get("medium", "")
        si_ver = self.ds.config["setup"].get("software version", "")
        si_ver = si_ver.strip("dev")  # for e.g. "2.2.1.0dev"
        if (medium == "CellCarrierB"
                and si_ver in ["2.2.1.0", "2.2.2.0", "2.2.2.1", "2.2.2.2",
                               "2.2.2.3",
                               # The issue can still occur for these versions
                               # of Shape-In if a bad configuration file
                               # is used.
                               "2.2.2.4", "2.3.0.0"]):
            cues.append(ICue(
                msg="Metadata: Please verify that 'medium' is really "
                    + "'CellCarrierB' (Shape-In issue #3)",
                level="alert",
                category="metadata wrong",
                identifier="Shape-In issue #3",
                cfg_section="setup",
                cfg_key="medium"))
        return cues

    def sanity_check(self):
        """Sanity check that tests whether the data can be accessed"""
        cues = []
        cues += self.check_feature_size()
        cues += self.check_metadata_bad()
        for cue in cues:
            cue.msg = "Sanity check failed: " + cue.msg
        return [ci for ci in cues if ci.level == "violation"]


def check_dataset(path_or_ds):
    """Check whether a dataset is complete

    Parameters
    ----------
    path_or_ds: str or pathlib.Path or RTDCBase
        Full path to a dataset on disk or an instance of RTDCBase

    Returns
    -------
    violations: list of str
        Dataset format violations (hard)
    alerts: list of str
        Dataset format alerts (soft)
    info: list of str
        Dataset information
    """
    aler = []
    info = []
    viol = []
    with IntegrityChecker(path_or_ds) as ic:
        # perform all checks
        icues = ic.check(expand_section=False)
        for cue in icues:
            if cue.level == "info":
                info.append(cue.msg)
            elif cue.level == "alert":
                aler.append(cue.msg)
            elif cue.level == "violation":
                viol.append(cue.msg)
    return sorted(viol), sorted(aler), sorted(info)



"""DCOR client interface"""
from functools import lru_cache
import json
import numbers
import time
import uuid

import numpy as np

from .. import definitions as dfn
from ..util import hashobj

from .config import Configuration
from .core import RTDCBase

try:
    import requests
except ModuleNotFoundError:
    REQUESTS_AVAILABLE = False
else:
    REQUESTS_AVAILABLE = True


class DCORAccessError(BaseException):
    pass


class APIHandler:
    """Handles the DCOR api with caching for simple queries"""
    #: these are cached to minimize network usage
    cache_queries = ["metadata", "size", "feature_list", "valid"]
    #: DCOR API Keys in the current session
    api_keys = []

    def __init__(self, url, api_key=""):
        self.url = url
        self.api_key = api_key
        self._cache = {}

    @classmethod
    def add_api_key(cls, api_key):
        """Add an API Key to the base class

        When accessing the DCOR API, all available API Keys are
        used to access a resource (trial and error).
        """
        if api_key.strip() and api_key not in APIHandler.api_keys:
            APIHandler.api_keys.append(api_key)

    def _get(self, query, feat=None, trace=None, event=None, api_key="",
             retries=3):
        qstr = "&query={}".format(query)
        if feat is not None:
            qstr += "&feature={}".format(feat)
        if trace is not None:
            qstr += "&trace={}".format(trace)
        if event is not None:
            qstr += "&event={}".format(event)
        apicall = self.url + qstr
        for _ in range(retries):
            req = requests.get(apicall, headers={"Authorization": api_key})
            try:
                jreq = req.json()
            except json.decoder.JSONDecodeError:
                time.sleep(0.1)  # wait a bit, maybe the server is overloaded
                continue
            else:
                break
        else:
            raise DCORAccessError("Could not complete query '{}', because "
                                  "the response did not contain any JSON-"
                                  "parseable data. Retried {} times.".format(
                                      apicall, retries))
        return jreq

    def get(self, query, feat=None, trace=None, event=None):
        if query in APIHandler.cache_queries and query in self._cache:
            result = self._cache[query]
        else:
            for api_key in [self.api_key] + APIHandler.api_keys:
                req = self._get(query, feat, trace, event, api_key)
                if req["success"]:
                    self.api_key = api_key  # remember working key
                    break
            else:
                raise DCORAccessError("Cannot access {}: {}".format(
                    query, req["error"]["message"]))
            result = req["result"]
            if query in APIHandler.cache_queries:
                self._cache[query] = result
        return result


class DCORNonScalarFeature:
    """Helper class for accessing non-scalar features"""

    def __init__(self, feat, api, size):
        self.identifier = api.url + ":" + feat  # for caching ancillaries
        self.feat = feat
        self.api = api
        self._size = size

    def __iter__(self):
        for idx in range(len(self)):
            yield self[idx]

    def __getitem__(self, event):
        if not isinstance(event, numbers.Integral):
            # slicing!
            indices = np.arange(len(self))[event]
            trace0 = self._get_item(indices[0])
            # determine the correct shape from the first feature
            oshape = [len(indices)] + list(trace0.shape)
            output = np.zeros(oshape, dtype=trace0.dtype)
            # populate the output array
            for ii, evid in enumerate(indices):
                output[ii] = self._get_item(evid)
            return output
        else:
            return self._get_item(event)

    def __len__(self):
        return self._size

    @lru_cache(maxsize=100)
    def _get_item(self, event):
        data = self.api.get(query="feature", feat=self.feat, event=event)
        return np.asarray(data)


class DCORContourFeature(DCORNonScalarFeature):
    """Helper class for accessing contour data"""

    def __init__(self, feat, api, size):
        super(DCORContourFeature, self).__init__(feat, api, size)
        self.shape = (size, np.nan, 2)

    def __getitem__(self, event):
        if not isinstance(event, numbers.Integral):
            # We cannot use the original method, because contours
            # may have different sizes! So we return a list.
            indices = np.arange(len(self))[event]
            output = []
            # populate the output list
            for evid in indices:
                output.append(self._get_item(evid))
            return output
        else:
            return self._get_item(event)


class DCORImageFeature(DCORNonScalarFeature):
    """Helper class for accessing image data"""

    def __init__(self, feat, api, size):
        super(DCORImageFeature, self).__init__(feat, api, size)
        metadata = self.api.get(query="metadata")
        self.shape = (size,
                      metadata["imaging"]["roi size y"],
                      metadata["imaging"]["roi size x"])


class DCORTraceItem(DCORNonScalarFeature):
    """Helper class for accessing traces"""
    def __init__(self, feat, api, size, samples_per_event):
        super(DCORTraceItem, self).__init__(feat, api, size)
        self.shape = (size, samples_per_event)

    @lru_cache(maxsize=100)
    def _get_item(self, event):
        data = self.api.get(query="feature", feat="trace",
                            trace=self.feat, event=event)
        return np.asarray(data)


class DCORTraceFeature:
    """Helper class for accessing traces"""

    def __init__(self, api, size):
        self.identifier = api.url + ":traces"
        self.api = api
        self._size = size
        metadata = self.api.get(query="metadata")
        self._samples_per_event = metadata["fluorescence"]["samples per event"]
        self.traces = api.get(query="trace_list")
        self._trace_objects = {}

        self.shape = (len(self.traces),
                      size,
                      self._samples_per_event
                      )

    def __contains__(self, key):
        return key in self.traces

    def __getitem__(self, trace):
        if trace in self.traces:
            if trace not in self._trace_objects:
                self._trace_objects[trace] = DCORTraceItem(
                    trace, self.api, self._size, self._samples_per_event)
            return self._trace_objects[trace]
        else:
            raise KeyError("trace '{}' not found!".format(trace))

    def __len__(self):
        return len(self.traces)

    def keys(self):
        return self.traces


class FeatureCache:
    """Download and cache (scalar only) features from DCOR"""

    def __init__(self, api, size):
        self.api = api
        self._features = self.api.get(query="feature_list")
        self._size = size
        self._scalar_cache = {}
        self._nonsc_features = {}

    def __contains__(self, key):
        return key in self._features

    def __getitem__(self, key):
        # user-level checking is done in core.py
        assert dfn.feature_exists(key)
        if key not in self._features:
            raise KeyError("Feature '{}' not found!".format(key))

        if key in self._scalar_cache:
            return self._scalar_cache[key]
        elif dfn.scalar_feature_exists(key):
            # download the feature and cache it
            feat = np.asarray(self.api.get(query="feature", feat=key))
            self._scalar_cache[key] = feat
            return feat
        elif key == "contour":
            if key not in self._nonsc_features:
                self._nonsc_features[key] = DCORContourFeature(key, self.api,
                                                               self._size)
            return self._nonsc_features[key]
        elif key == "trace":
            if "trace" not in self._nonsc_features:
                self._nonsc_features["trace"] = DCORTraceFeature(self.api,
                                                                 self._size)
            return self._nonsc_features["trace"]
        elif key in ["image", "mask"]:
            self._nonsc_features[key] = DCORImageFeature(key, self.api,
                                                         self._size)
            return self._nonsc_features[key]
        else:
            raise ValueError(f"No DCOR handler for feature '{key}'!")

    def __iter__(self):
        # dict-like behavior
        for key in self.keys():
            yield key

    def keys(self):
        return self._features


class RTDC_DCOR(RTDCBase):
    def __init__(self, url, use_ssl=None, host="dcor.mpl.mpg.de",
                 api_key="", *args, **kwargs):
        """Wrap around the DCOR API

        Parameters
        ----------
        url: str
            Full URL or resource identifier; valid values are

            - `<https://dcor.mpl.mpg.de/api/3/action/dcserv?id=
              b1404eb5-f661-4920-be79-5ff4e85915d5>`_
            - dcor.mpl.mpg.de/api/3/action/dcserv?id=b1404eb5-f
              661-4920-be79-5ff4e85915d5
            - b1404eb5-f661-4920-be79-5ff4e85915d5
        use_ssl: bool
            Set this to False to disable SSL (should only be used for
            testing). Defaults to None (does not force SSL if the URL
            starts with "http://").
        host: str
            The host machine (used if the host is not given in `url`)
        api_key: str
            API key to access private resources
        *args:
            Arguments for `RTDCBase`
        **kwargs:
            Keyword arguments for `RTDCBase`

        Attributes
        ----------
        path: str
            Full URL to the DCOR resource
        """
        if not REQUESTS_AVAILABLE:
            raise ModuleNotFoundError(
                "Package `requests` required for DCOR format!")

        super(RTDC_DCOR, self).__init__(*args, **kwargs)

        self._hash = None
        self.path = RTDC_DCOR.get_full_url(url, use_ssl, host)
        self.api = APIHandler(url=self.path, api_key=api_key)

        # Parse configuration
        self.config = Configuration(cfg=self.api.get(query="metadata"))

        # Get size
        self._size = int(self.api.get(query="size"))

        # Setup events
        self._events = FeatureCache(self.api, size=self._size)

        # Override logs property with HDF5 data
        self.logs = {}

        self.title = "{} - M{}".format(self.config["experiment"]["sample"],
                                       self.config["experiment"]["run index"])

        # Set up filtering
        self._init_filters()

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        pass

    def __len__(self):
        return self._size

    @staticmethod
    def get_full_url(url, use_ssl, host):
        """Return the full URL to a DCOR resource

        Parameters
        ----------
        url: str
            Full URL or resource identifier; valid values are

            - https://dcor.mpl.mpg.de/api/3/action/dcserv?id=caab96f6-
              df12-4299-aa2e-089e390aafd5'
            - dcor.mpl.mpg.de/api/3/action/dcserv?id=caab96f6-df12-
              4299-aa2e-089e390aafd5
            - caab96f6-df12-4299-aa2e-089e390aafd5
        use_ssl: bool
            Set this to False to disable SSL (should only be used for
            testing). Defaults to None (does not force SSL if the URL
            starts with "http://").
        host: str
            Use this host if it is not specified in `url`
        """
        if use_ssl is None:
            if url.startswith("http://"):
                # user wanted it that way
                web = "http"
            else:
                web = "https"
        elif use_ssl:
            web = "https"
        else:
            web = "http"
        if url.count("://"):
            base = url.split("://", 1)[1]
        else:
            base = url
        if base.count("/"):
            host, api = base.split("/", 1)
        else:
            api = "api/3/action/dcserv?id=" + base
        new_url = "{}://{}/{}".format(web, host, api)
        return new_url

    @property
    def hash(self):
        """Hash value based on file name and content"""
        if self._hash is None:
            tohash = [self.path]
            self._hash = hashobj(tohash)
        return self._hash


def is_dcor_url(string):
    if not isinstance(string, str):
        return False
    elif is_uuid(string):
        return True
    else:  # we have a string
        if string.startswith("http://") or string.startswith("https://"):
            return True  # pretty safe bet
        elif string.count("/api/3/action/dcserv?id="):
            return True  # not so safe, but highly improbable folder structure
        else:
            return False


def is_uuid(string):
    try:
        uuid_obj = uuid.UUID(string)
    except ValueError:
        return False
    return str(uuid_obj) == string



import pathlib
import tempfile

from dclab import util
import pytest


@pytest.fixture(autouse=True)
def clear_cache():
    util._hashfile_cached.cache_clear()


def test_hashfile_basic():
    td = pathlib.Path(tempfile.mkdtemp("test_hashfile"))
    p1 = td / "test_1.txt"
    p1.write_text("Lorem ipsum")
    util.hashfile(p1)
    util.hashfile(p1)
    assert util._hashfile_cached.cache_info().misses == 1
    assert util._hashfile_cached.cache_info().hits == 1
    assert util._hashfile_cached.cache_info().currsize == 1

    p2 = td / "test_2.txt"
    p2.write_text("dolor sit amet.")
    util.hashfile(p2)
    assert util._hashfile_cached.cache_info().misses == 2
    assert util._hashfile_cached.cache_info().hits == 1
    assert util._hashfile_cached.cache_info().currsize == 2


def test_hashfile_modified():
    td = pathlib.Path(tempfile.mkdtemp("test_hashfile"))
    p1 = td / "test_1.txt"
    p1.write_text("Lorem ipsum")
    util.hashfile(p1)
    util.hashfile(p1)
    assert util._hashfile_cached.cache_info().misses == 1
    assert util._hashfile_cached.cache_info().hits == 1
    assert util._hashfile_cached.cache_info().currsize == 1

    p1.write_text("dolor sit amet.")
    util.hashfile(p1)
    assert util._hashfile_cached.cache_info().misses == 2
    assert util._hashfile_cached.cache_info().hits == 1
    assert util._hashfile_cached.cache_info().currsize == 2


def test_hashfile_modified_quickly():
    td = pathlib.Path(tempfile.mkdtemp("test_hashfile"))
    p1 = td / "test_1.txt"
    p1.write_text("Lorem ipsum")
    util.hashfile(p1)
    p1.write_text("dolor sit amet.")
    util.hashfile(p1)
    assert util._hashfile_cached.cache_info().misses == 2
    assert util._hashfile_cached.cache_info().hits == 0
    assert util._hashfile_cached.cache_info().currsize == 2



"""Test machine learning tools with tensorflow"""
import pathlib
import tempfile
from unittest import mock

import numpy as np
import pytest

import dclab
from dclab import new_dataset
from dclab.rtdc_dataset import feat_anc_ml
from dclab.rtdc_dataset.feat_anc_ml import hook_tensorflow

from helper_methods import example_data_dict, retrieve_data

tf = pytest.importorskip("tensorflow")


data_dir = pathlib.Path(__file__).parent / "data"


@pytest.fixture(autouse=True)
def cleanup_plugin_features():
    """Fixture used to cleanup plugin feature tests"""
    # code run before the test
    pass
    # then the test is run
    yield
    # code run after the test
    # remove our test plugin examples
    feat_anc_ml.remove_all_ml_features()


def make_data(add_feats=None, sizes=None):
    if sizes is None:
        sizes = [100, 130]
    if add_feats is None:
        add_feats = ["area_um", "deform"]
    keys = add_feats + ["time", "frame", "fl3_width"]
    data = []
    for size in sizes:
        data.append(new_dataset(example_data_dict(size=size, keys=keys)))
    return data


def make_bare_model(ml_feats=None):
    if ml_feats is None:
        ml_feats = ["area_um", "deform"]
    dc_data = make_data(ml_feats)
    # obtain train and test datasets
    tfdata = hook_tensorflow.assemble_tf_dataset_scalars(
        dc_data=dc_data,
        labels=[0, 1],
        feature_inputs=ml_feats)
    return standard_model(tfdata)


def standard_model(tfdata, epochs=1, final_size=2):
    # build the model
    num_feats = tfdata.as_numpy_iterator().next()[0].shape[1]
    model = tf.keras.Sequential(
        layers=[
            tf.keras.layers.Input(shape=(num_feats,)),
            tf.keras.layers.Dense(128, activation='relu'),
            tf.keras.layers.Dense(32),
            tf.keras.layers.Dropout(0.2),
            tf.keras.layers.Dense(final_size)
        ],
        name="scalar_features"
    )

    # fit the model to the training data
    if final_size == 1:
        loss_fn = tf.keras.losses.BinaryCrossentropy(from_logits=True)
    else:
        loss_fn = tf.keras.losses.SparseCategoricalCrossentropy(
            from_logits=True)
    model.compile(optimizer='adam', loss=loss_fn, metrics=['accuracy'])
    model.fit(tfdata, epochs=epochs)

    return model


def test_af_ml_class_changed_model():
    # setup
    dict1 = example_data_dict(size=1000, keys=["area_um", "aspect", "fl1_max"])
    dict1["deform"] = np.linspace(.01, .02, 1000)
    dict2 = example_data_dict(size=1000, keys=["area_um", "aspect", "fl1_max"])
    dict2["deform"] = np.linspace(.03, .04, 1000)
    ds1 = dclab.new_dataset(dict1)
    ds2 = dclab.new_dataset(dict2)
    tfdata = hook_tensorflow.assemble_tf_dataset_scalars(
        dc_data=[ds1, ds2],
        labels=[0, 1],
        feature_inputs=["deform"])
    bare_model1 = standard_model(tfdata, epochs=10)
    dc_model1 = hook_tensorflow.TensorflowModel(
        bare_model=bare_model1,
        inputs=["deform"],
        outputs=["ml_score_001", "ml_score_002"])

    bare_model2 = standard_model(tfdata, epochs=10)
    dc_model2 = hook_tensorflow.TensorflowModel(
        bare_model=bare_model2,
        inputs=["deform"],
        # flip feature 1 and 2
        outputs=["ml_score_002", "ml_score_001"])

    feat_anc_ml.MachineLearningFeature(feature_name="ml_score_001",
                                       dc_model=dc_model1)
    feat_anc_ml.MachineLearningFeature(feature_name="ml_score_002",
                                       dc_model=dc_model1)

    # checkout the ml_class feature
    assert "ml_score_001" in ds1
    assert np.all(ds1["ml_class"] == 0)

    feat_anc_ml.ml_feature.remove_all_ml_features()

    feat_anc_ml.MachineLearningFeature(feature_name="ml_score_001",
                                       dc_model=dc_model2)
    feat_anc_ml.MachineLearningFeature(feature_name="ml_score_002",
                                       dc_model=dc_model2)

    assert "ml_score_001" in ds1
    assert np.all(ds1["ml_class"] == 1)


def test_af_ml_score_basic():
    """Slight modification of related test in test_ml.py using ancillaries"""
    # setup
    dict1 = example_data_dict(size=1000, keys=["area_um", "aspect", "fl1_max"])
    dict1["deform"] = np.linspace(.01, .02, 1000)
    dict2 = example_data_dict(size=1000, keys=["area_um", "aspect", "fl1_max"])
    dict2["deform"] = np.linspace(.02, .03, 1000)
    ds1 = dclab.new_dataset(dict1)
    ds2 = dclab.new_dataset(dict2)
    tfdata = hook_tensorflow.assemble_tf_dataset_scalars(
        dc_data=[ds1, ds2],
        labels=[0, 1],
        feature_inputs=["deform"])
    bare_model = standard_model(tfdata, epochs=10)
    # test dataset
    dictt = example_data_dict(size=6, keys=["area_um", "aspect", "fl1_max"])
    dictt["deform"] = np.linspace(.015, .025, 6, endpoint=True)
    dst = dclab.new_dataset(dictt)
    # DC model
    dc_model = hook_tensorflow.TensorflowModel(
        bare_model=bare_model,
        inputs=["deform"],
        outputs=["ml_score_low", "ml_score_hig"])
    # register
    feat_anc_ml.MachineLearningFeature(feature_name="ml_score_low",
                                       dc_model=dc_model)
    feat_anc_ml.MachineLearningFeature(feature_name="ml_score_hig",
                                       dc_model=dc_model)
    # get the class from the predicted values
    low = dst["ml_score_low"]
    hig = dst["ml_score_hig"]
    hl = np.concatenate((low.reshape(-1, 1), hig.reshape(-1, 1)), axis=1)
    actual = np.argmax(hl, axis=1)
    # these are the expected classes
    expected = dictt["deform"] > 0.02
    # sanity check
    assert np.sum(expected) == np.sum(~expected)
    # actual verification of test case
    assert np.all(expected == actual)


def test_af_ml_score_label_registered():
    """Test whether the correct label is returned"""
    ml_feats = ["area_um", "deform"]
    bare_model = make_bare_model(ml_feats=ml_feats)
    model = hook_tensorflow.TensorflowModel(
        bare_model=bare_model,
        inputs=ml_feats,
        outputs=["ml_score_low", "ml_score_hig"],
        info={"output labels": ["Low label", "High label"]})

    # register the features
    feat_anc_ml.MachineLearningFeature(feature_name="ml_score_low",
                                       dc_model=model)
    feat_anc_ml.MachineLearningFeature(feature_name="ml_score_hig",
                                       dc_model=model)

    # get labels
    label1 = dclab.dfn.get_feature_label("ml_score_low")
    label2 = dclab.dfn.get_feature_label("ml_score_hig")
    assert label1 == "Low label"
    assert label2 == "High label"


def test_af_ml_score_registration_sanity_checks():
    ml_feats = ["area_um", "deform"]
    dc_model = hook_tensorflow.TensorflowModel(
        bare_model=make_bare_model(),
        inputs=ml_feats,
        outputs=["ml_score_abc", "ml_score_cde"]
    )
    dictt = example_data_dict(size=6, keys=["area_um", "aspect", "fl1_max"])
    dictt["deform"] = np.linspace(.015, .025, 6, endpoint=True)
    dst = dclab.new_dataset(dictt)
    assert "ml_score_abc" not in dst, "before model registration"
    feat_anc_ml.MachineLearningFeature(feature_name="ml_score_abc",
                                       dc_model=dc_model)
    assert "ml_score_abc" in dst, "after model registration"
    feat_anc_ml.remove_all_ml_features()
    assert "ml_score_abc" not in dst, "after model unregistration"


def test_af_ml_score_register_same_output_feature_should_fail():
    ml_feats = ["area_um", "deform"]
    dc_model = hook_tensorflow.TensorflowModel(
        bare_model=make_bare_model(),
        inputs=ml_feats,
        outputs=["ml_score_abc", "ml_score_cde"]
    )
    feat_anc_ml.MachineLearningFeature(feature_name="ml_score_abc",
                                       dc_model=dc_model)

    dc_model2 = hook_tensorflow.TensorflowModel(
        bare_model=make_bare_model(),
        inputs=ml_feats,
        outputs=["ml_score_abc", "ml_score_rge"]
    )

    with pytest.raises(ValueError,
                       match="Cannot register two MachineLearningFeatures"):
        feat_anc_ml.MachineLearningFeature(feature_name="ml_score_abc",
                                           dc_model=dc_model2)


def test_assemble_tf_dataset_scalars_shuffle():
    ml_feats = ["deform", "area_um"]
    dc_data = make_data(ml_feats)
    # create a shuffled dataset
    tfdata = hook_tensorflow.assemble_tf_dataset_scalars(
        dc_data=dc_data,
        labels=[0, 1],
        feature_inputs=ml_feats,
        shuffle=True
    )
    # reproduce the shuffling
    area_um = hook_tensorflow.shuffle_array(
        np.concatenate([ds["area_um"] for ds in dc_data]))
    deform = hook_tensorflow.shuffle_array(
        np.concatenate([ds["deform"] for ds in dc_data]))
    # get the actual data
    tffeats = np.concatenate([dd for dd, ll in tfdata], axis=0)
    actual_deform = tffeats[:, 0]
    actual_area_um = tffeats[:, 1]
    assert np.all(np.array(deform, dtype=np.float32) == actual_deform)
    assert np.all(np.array(area_um, dtype=np.float32) == actual_area_um)


def test_assemble_tf_dataset_scalars_split():
    ml_feats = ["deform", "area_um"]
    dc_data = make_data(ml_feats)
    # create a shuffled dataset
    tfdata = hook_tensorflow.assemble_tf_dataset_scalars(
        dc_data=dc_data,
        labels=[0, 1],
        feature_inputs=ml_feats,
        shuffle=True
    )
    tfsplit1, tfsplit2 = hook_tensorflow.assemble_tf_dataset_scalars(
        dc_data=dc_data,
        labels=[0, 1],
        feature_inputs=ml_feats,
        split=.8,
        shuffle=True,
    )
    # reproduce the splitting
    tffeats = np.concatenate([dd for dd, ll in tfdata], axis=0)
    s1feats = np.concatenate([dd for dd, ll in tfsplit1], axis=0)
    s2feats = np.concatenate([dd for dd, ll in tfsplit2], axis=0)
    assert np.all(tffeats[:len(s1feats)] == s1feats)
    assert np.all(tffeats[len(s1feats):] == s2feats)


def test_assemble_tf_dataset_scalars_split_bad():
    ml_feats = ["deform", "area_um"]
    dc_data = make_data(ml_feats)
    try:
        hook_tensorflow.assemble_tf_dataset_scalars(
            dc_data=dc_data,
            labels=[0, 1],
            feature_inputs=ml_feats,
            split=43,
            shuffle=True)
    except ValueError:
        pass
    else:
        assert False, "Invalid split parameter"


def test_assemble_tf_dataset_scalars_non_scalar():
    ml_feats = ["deform", "area_um", "image"]
    dc_data = make_data(ml_feats)
    # create a shuffled dataset
    try:
        hook_tensorflow.assemble_tf_dataset_scalars(
            dc_data=dc_data,
            labels=[0, 1],
            feature_inputs=ml_feats)
    except ValueError:
        pass
    else:
        assert False, "'image' should not be supported for scalar dataset"


def test_basic_inference():
    # setup
    dict1 = example_data_dict(size=1000, keys=["area_um", "aspect", "fl1_max"])
    dict1["deform"] = np.linspace(.01, .02, 1000)
    dict2 = example_data_dict(size=1000, keys=["area_um", "aspect", "fl1_max"])
    dict2["deform"] = np.linspace(.02, .03, 1000)
    ds1 = new_dataset(dict1)
    ds2 = new_dataset(dict2)
    tfdata = hook_tensorflow.assemble_tf_dataset_scalars(
        dc_data=[ds1, ds2],
        labels=[0, 1],
        feature_inputs=["deform"])
    bare_model = standard_model(tfdata, epochs=10)
    # test dataset
    dictt = example_data_dict(size=6, keys=["area_um", "aspect", "fl1_max"])
    dictt["deform"] = np.linspace(.015, .025, 6, endpoint=True)
    dst = new_dataset(dictt)
    # DC model
    model = hook_tensorflow.TensorflowModel(
        bare_model=bare_model,
        inputs=["deform"],
        outputs=["ml_score_low", "ml_score_hig"])
    scores = model.predict(dst)
    # get the class from the predicted values
    low = scores["ml_score_low"]
    hig = scores["ml_score_hig"]
    hl = np.concatenate((low.reshape(-1, 1), hig.reshape(-1, 1)), axis=1)
    actual = np.argmax(hl, axis=1)
    # these are the expected classes
    expected = dictt["deform"] > 0.02
    # sanity check
    assert np.sum(expected) == np.sum(~expected)
    # actual verification of test case
    assert np.all(expected == actual)


def test_basic_inference_one_output():
    # setup
    dict1 = example_data_dict(size=1000, keys=["area_um", "aspect", "fl1_max"])
    dict1["deform"] = np.linspace(.01, .02, 1000)
    dict2 = example_data_dict(size=1000, keys=["area_um", "aspect", "fl1_max"])
    dict2["deform"] = np.linspace(.02, .03, 1000)
    ds1 = new_dataset(dict1)
    ds2 = new_dataset(dict2)
    tfdata = hook_tensorflow.assemble_tf_dataset_scalars(
        dc_data=[ds1, ds2],
        labels=[0, 1],
        feature_inputs=["deform"])
    bare_model = standard_model(tfdata, epochs=10, final_size=1)
    # test dataset
    dictt = example_data_dict(size=6, keys=["area_um", "aspect", "fl1_max"])
    dictt["deform"] = np.linspace(.015, .025, 6, endpoint=True)
    dst = new_dataset(dictt)
    # DC model
    model = hook_tensorflow.TensorflowModel(
        bare_model=bare_model,
        inputs=["deform"],
        outputs=["ml_score_loh"])
    scores = model.predict(dst)
    # get the class from the predicted values
    actual = scores["ml_score_loh"] > 0.5
    # these are the expected classes
    expected = dictt["deform"] > 0.02
    # sanity check
    assert np.sum(expected) == np.sum(~expected)
    # actual verification of test case
    assert np.all(expected == actual)


def test_get_dataset_event_feature():
    ml_feats = ["deform", "area_um"]
    dc_data = make_data(ml_feats)
    # create a shuffled dataset
    tfdata = hook_tensorflow.assemble_tf_dataset_scalars(
        dc_data=dc_data,
        labels=[0, 1],
        feature_inputs=ml_feats,
        shuffle=True
    )
    # get the actual data
    tffeats = np.concatenate([dd for dd, ll in tfdata], axis=0)
    actual_deform = tffeats[:, 0]
    actual_area_um = tffeats[:, 1]
    event_index = 42
    event_deform = hook_tensorflow.get_dataset_event_feature(
        dc_data=dc_data,
        feature="deform",
        tf_dataset_indices=[event_index],
        shuffle=True
    )[0]
    assert actual_deform[event_index] == np.float32(event_deform)
    event_area_um = hook_tensorflow.get_dataset_event_feature(
        dc_data=dc_data,
        feature="area_um",
        tf_dataset_indices=[event_index],
        shuffle=True
    )
    assert actual_area_um[event_index] == np.float32(event_area_um)


def test_get_dataset_event_feature_bad_index():
    ml_feats = ["deform", "area_um"]
    dc_data = make_data(ml_feats)
    try:
        hook_tensorflow.get_dataset_event_feature(
            dc_data=dc_data,
            feature="area_um",
            tf_dataset_indices=[10 * np.sum([len(ds) for ds in dc_data])],
            shuffle=True
        )
    except IndexError:
        pass
    else:
        assert False, "Event index too large"


def test_get_dataset_event_feature_no_tf_indices():
    ml_feats = ["deform", "area_um"]
    dc_data = make_data(ml_feats)
    indices0 = hook_tensorflow.get_dataset_event_feature(
        dc_data=dc_data,
        feature="index",
        dc_data_indices=[0],
        shuffle=False
    )
    assert list(range(1, len(dc_data[0]) + 1)) == indices0

    indices1 = hook_tensorflow.get_dataset_event_feature(
        dc_data=dc_data,
        feature="index",
        dc_data_indices=[1],
        shuffle=False
    )
    assert list(range(1, len(dc_data[1]) + 1)) == indices1


def test_get_dataset_event_feature_split():
    ml_feats = ["deform", "area_um"]
    dc_data = make_data(ml_feats)
    tfsplit1, tfsplit2 = hook_tensorflow.assemble_tf_dataset_scalars(
        dc_data=dc_data,
        labels=[0, 1],
        feature_inputs=ml_feats,
        split=.8,
        shuffle=True,
    )
    event_index = 5
    event_area_um = hook_tensorflow.get_dataset_event_feature(
        dc_data=dc_data,
        feature="area_um",
        tf_dataset_indices=[event_index],
        split=.8,
        split_index=0,
        shuffle=True
    )
    s1feats = np.concatenate([dd for dd, ll in tfsplit1], axis=0)
    assert s1feats[event_index, 1] == np.float32(event_area_um)
    event_area_um2 = hook_tensorflow.get_dataset_event_feature(
        dc_data=dc_data,
        feature="area_um",
        tf_dataset_indices=[event_index],
        split=.8,
        split_index=1,
        shuffle=True
    )
    s2feats = np.concatenate([dd for dd, ll in tfsplit2], axis=0)
    assert s2feats[event_index, 1] == np.float32(event_area_um2)


def test_get_dataset_event_feature_split_bad():
    ml_feats = ["deform", "area_um"]
    dc_data = make_data(ml_feats)
    try:
        hook_tensorflow.get_dataset_event_feature(
            dc_data=dc_data,
            feature="area_um",
            tf_dataset_indices=[1],
            split=-1,
            split_index=1,
            shuffle=True
        )
    except ValueError:
        pass
    else:
        assert False, "Invalid split parameter"


def test_get_dataset_event_feature_split_bad2():
    ml_feats = ["deform", "area_um"]
    dc_data = make_data(ml_feats)
    try:
        hook_tensorflow.get_dataset_event_feature(
            dc_data=dc_data,
            feature="area_um",
            tf_dataset_indices=[1],
            split=0,
            split_index=1,
            shuffle=True
        )
    except IndexError:
        pass
    else:
        assert False, "Invalid split parameter"


def test_model_tensorflow_has_sigmoid_activation_1():
    ml_feats = ["deform", "area_um"]
    stock_model = make_bare_model(ml_feats)
    bare_model = tf.keras.Sequential([stock_model, tf.keras.layers.Dense(1)])
    dc_model = hook_tensorflow.TensorflowModel(bare_model=bare_model,
                                               inputs=ml_feats,
                                               outputs=["ml_score_tst"])
    assert not dc_model.has_sigmoid_activation()


def test_model_tensorflow_has_sigmoid_activation_2():
    ml_feats = ["deform", "area_um"]
    stock_model = make_bare_model(ml_feats)
    bare_model = tf.keras.Sequential(
        [stock_model, tf.keras.layers.Dense(1, activation="sigmoid")])
    dc_model = hook_tensorflow.TensorflowModel(bare_model=bare_model,
                                               inputs=ml_feats,
                                               outputs=["ml_score_tst"])
    assert dc_model.has_sigmoid_activation()


def test_model_tensorflow_has_softmax_layer_1():
    ml_feats = ["deform", "area_um"]
    stock_model = make_bare_model(ml_feats)
    bare_model = tf.keras.Sequential(
        [stock_model, tf.keras.layers.Dense(2)])
    dc_model = hook_tensorflow.TensorflowModel(bare_model=bare_model,
                                               inputs=ml_feats,
                                               outputs=["ml_score_tst",
                                                        "ml_score_ts2"])
    assert not dc_model.has_softmax_layer()


def test_model_tensorflow_has_softmax_layer_2():
    ml_feats = ["deform", "area_um"]
    stock_model = make_bare_model(ml_feats)
    bare_model = tf.keras.Sequential(
        [stock_model, tf.keras.layers.Dense(2), tf.keras.layers.Softmax()])
    dc_model = hook_tensorflow.TensorflowModel(bare_model=bare_model,
                                               inputs=ml_feats,
                                               outputs=["ml_score_tst",
                                                        "ml_score_ts2"])
    assert dc_model.has_softmax_layer()


def test_modc_apply_model():
    feat_anc_ml.load_ml_feature(data_dir / "feat_anc_ml_tf_rbc.modc")
    path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    with dclab.new_dataset(path) as ds:
        assert np.sum(ds["ml_score_rbc"] < .5) == 12
        assert np.sum(ds["ml_score_rbc"] >= .5) == 6
        assert np.allclose(ds["ml_score_rbc"][0], 0.21917653)


def test_modc_export_model_bad_path_1():
    tmpdir = pathlib.Path(tempfile.mkdtemp(prefix="dclab_ml"))
    bare_model = make_bare_model()
    afile = tmpdir / "afile"
    afile.write_text("test")
    with pytest.raises(ValueError,
                       match=r"Output `path` should be a directory"):
        feat_anc_ml.modc.export_model(path=afile,
                                      model=bare_model,
                                      )


def test_modc_export_model_bad_path_2():
    tmpdir = pathlib.Path(tempfile.mkdtemp(prefix="dclab_ml"))
    bare_model = make_bare_model()
    afile = tmpdir / "afile"
    afile.write_text("test")
    with pytest.raises(
            ValueError,
            match="Model output directory should be empty"):
        feat_anc_ml.modc.export_model(path=tmpdir,
                                      model=bare_model,
                                      )


def test_modc_export_model_enforce_format():
    tmpdir = pathlib.Path(tempfile.mkdtemp(prefix="dclab_ml"))
    bare_model = make_bare_model()
    feat_anc_ml.modc.export_model(path=tmpdir,
                                  model=bare_model,
                                  enforce_formats=["tensorflow-SavedModel"]
                                  )


def test_modc_export_model_enforce_format_bad_format():
    tmpdir = pathlib.Path(tempfile.mkdtemp(prefix="dclab_ml"))
    bare_model = make_bare_model()
    with pytest.raises(
            ValueError,
            match="Unsupported format 'library-format-does-not-exist'"):
        feat_anc_ml.modc.export_model(path=tmpdir,
                                      model=bare_model,
                                      enforce_formats=[
                                          "library-format-does-not-exist"]
                                      )


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.feat_anc_ml.modc.'
                            + 'ModelFormatExportFailedWarning')
def test_modc_export_model_enforce_format_bad_model():
    tmpdir = pathlib.Path(tempfile.mkdtemp(prefix="dclab_ml"))
    with pytest.raises(ValueError,
                       match="Expected an object of type `Trackable`"
                             "|Expected a Trackable object for export"):
        feat_anc_ml.modc.export_model(path=tmpdir,
                                      model=mock.MagicMock(),
                                      enforce_formats=["tensorflow-SavedModel"]
                                      )


def test_modc_load_basic():
    # setup
    tmpdir = pathlib.Path(tempfile.mkdtemp(prefix="dclab_ml"))
    bare_model = make_bare_model()
    model = hook_tensorflow.TensorflowModel(bare_model=bare_model,
                                            inputs=["image"],
                                            outputs=["ml_score_tst"])
    pout = tmpdir / "test.modc"
    feat_anc_ml.save_modc(path=pout, dc_models=model)
    # assert
    feat_anc_ml.load_modc(path=pout)


def test_modc_load_as_ancillary_feature():
    # setup
    tmpdir = pathlib.Path(tempfile.mkdtemp(prefix="dclab_ml"))
    bare_model = make_bare_model()
    model = hook_tensorflow.TensorflowModel(bare_model=bare_model,
                                            inputs=["image"],
                                            outputs=["ml_score_tst"])
    pout = tmpdir / "test.modc"
    feat_anc_ml.save_modc(path=pout, dc_models=model)
    # assert
    aml_list = dclab.load_ml_feature(pout)
    assert len(aml_list) == 1
    assert aml_list[0].feature_name == "ml_score_tst"
    assert "ml_score_tst" in dclab.MachineLearningFeature.feature_names


def test_modc_load_bad_format():
    # setup
    tmpdir = pathlib.Path(tempfile.mkdtemp(prefix="dclab_ml"))
    bare_model = make_bare_model()
    model = hook_tensorflow.TensorflowModel(bare_model=bare_model,
                                            inputs=["image"],
                                            outputs=["ml_score_tst"])
    pout = tmpdir / "test.modc"
    feat_anc_ml.save_modc(path=pout, dc_models=model)
    # assert
    try:
        feat_anc_ml.load_modc(path=pout, from_format="format-does-not-exist")
    except ValueError:
        pass
    else:
        assert False, "bad format should not work"


def test_modc_save_basic():
    # setup
    tmpdir = pathlib.Path(tempfile.mkdtemp(prefix="dclab_ml"))
    bare_model = make_bare_model()
    model = hook_tensorflow.TensorflowModel(bare_model=bare_model,
                                            inputs=["deform"],
                                            outputs=["ml_score_tst"])
    pout = tmpdir / "test.modc"
    # assert
    feat_anc_ml.save_modc(path=pout, dc_models=model)


def test_modc_save_load_infer():
    # setup
    dict1 = example_data_dict(size=1000, keys=["area_um", "aspect", "fl1_max"])
    dict1["deform"] = np.linspace(.01, .02, 1000)
    dict2 = example_data_dict(size=1000, keys=["area_um", "aspect", "fl1_max"])
    dict2["deform"] = np.linspace(.02, .03, 1000)
    ds1 = new_dataset(dict1)
    ds2 = new_dataset(dict2)
    tfdata = hook_tensorflow.assemble_tf_dataset_scalars(
        dc_data=[ds1, ds2],
        labels=[0, 1],
        feature_inputs=["deform"])
    bare_model = standard_model(tfdata, epochs=10)
    # DC model
    model = hook_tensorflow.TensorflowModel(
        bare_model=bare_model,
        inputs=["deform"],
        outputs=["ml_score_low", "ml_score_hig"])
    # save model
    tmpdir = pathlib.Path(tempfile.mkdtemp(prefix="dclab_ml_modc"))
    pout = tmpdir / "test.modc"
    feat_anc_ml.save_modc(path=pout, dc_models=model)
    # load model
    model_loaded = feat_anc_ml.load_modc(path=pout)[0]
    # test dataset
    dictt = example_data_dict(size=6, keys=["area_um", "aspect", "fl1_max"])
    dictt["deform"] = np.linspace(.015, .025, 6, endpoint=True)
    dst = new_dataset(dictt)
    expected = model.predict(dst)
    actual = model_loaded.predict(dst)
    # assert
    assert np.all(expected["ml_score_low"] == actual["ml_score_low"])
    assert np.all(expected["ml_score_hig"] == actual["ml_score_hig"])


def test_models_get_dataset_features():
    ml_feats = ["area_um", "deform"]
    bare_model = make_bare_model(ml_feats=ml_feats)
    mod = hook_tensorflow.TensorflowModel(bare_model=bare_model,
                                          inputs=ml_feats,
                                          outputs=["ml_score_t01",
                                                   "ml_score_t01"])
    ds = make_data(add_feats=ml_feats, sizes=[42])[0]
    fdata = mod.get_dataset_features(ds, dtype=np.float32)
    assert np.all(fdata[:, 0] == np.array(ds["area_um"], dtype=np.float32))


def test_models_get_dataset_features_with_tensorlow():
    ml_feats = ["area_um", "deform"]
    bare_model = make_bare_model(ml_feats=ml_feats)
    mod = hook_tensorflow.TensorflowModel(bare_model=bare_model,
                                          inputs=ml_feats,
                                          outputs=["ml_score_t01",
                                                   "ml_score_t01"])
    ds = make_data(add_feats=ml_feats, sizes=[42])[0]
    fdata = mod.get_dataset_features(ds, dtype=np.float32)
    tfdata = hook_tensorflow.assemble_tf_dataset_scalars(
        dc_data=[ds],
        feature_inputs=ml_feats,
        shuffle=False
    )
    tf_area_um = np.concatenate([x[:, 0] for x in tfdata], axis=0)
    assert np.all(fdata[:, 0] == tf_area_um)


def test_models_prediction_runs_through():
    ml_feats = ["area_um", "deform"]
    bare_model = make_bare_model(ml_feats=ml_feats)
    mod = hook_tensorflow.TensorflowModel(bare_model=bare_model,
                                          inputs=ml_feats,
                                          outputs=["ml_score_t01",
                                                   "ml_score_t02"])
    ds = make_data(add_feats=ml_feats, sizes=[42])[0]
    out = mod.predict(ds)
    assert "ml_score_t01" in out
    assert "ml_score_t02" in out



"""Split an .rtdc file into smaller .rtdc files"""
import argparse
import pathlib
import warnings

from ..rtdc_dataset import fmt_tdms, new_dataset, RTDCWriter

from . import common


def split(path_in=None, path_out=None, split_events=10000,
          skip_initial_empty_image=True, skip_final_empty_image=True,
          ret_out_paths=False, verbose=False):
    """Split a measurement file

    Parameters
    ----------
    path_in: str or pathlib.Path
        Path of input measurement file
    path_out: str or pathlib.Path
        Path to output directory (optional)
    split_events: int
        Maximum number of events in each output file
    skip_initial_empty_image: bool
        Remove the first event of the dataset if the image is zero.
    skip_final_empty_image: bool
        Remove the final event of the dataset if the image is zero.
    ret_out_paths:
        If True, return the list of output file paths.
    verbose: bool
        If `True`, print messages to stdout

    Returns
    -------
    [out_paths]: list of pathlib.Path
        List of generated files (only if `ret_out_paths` is specified)
    """
    if path_in is None:
        parser = split_parser()
        args = parser.parse_args()

        path_in = pathlib.Path(args.path_in).resolve()
        path_out = args.path_out
        split_events = args.split_events
        skip_initial_empty_image = not args.include_empty_boundary_images
        skip_final_empty_image = not args.include_empty_boundary_images
        verbose = True

    if path_out in ["SAME", None]:  # default to input directory
        path_out = path_in.parent

    path_in = pathlib.Path(path_in)
    path_out = pathlib.Path(path_out)
    logs = {"dclab-split": common.get_command_log(paths=[path_in])}
    paths_gen = []
    paths_temp = []
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        # ignore ResourceWarning: unclosed file <_io.BufferedReader...>
        warnings.simplefilter("ignore", ResourceWarning)  # noqa: F821
        if fmt_tdms.NPTDMS_AVAILABLE:  # tdms-related warning filters
            # ignore SlowVideoWarning
            warnings.simplefilter("ignore",
                                  fmt_tdms.event_image.SlowVideoWarning)
            if skip_initial_empty_image:
                # If the initial frame is skipped when empty,
                # suppress any related warning messages.
                warnings.simplefilter(
                    "ignore",
                    fmt_tdms.event_image.InitialFrameMissingWarning)

        with new_dataset(path_in) as ds:
            for ll in ds.logs:
                logs[f"src-{ll}"] = ds.logs[ll]
            num_files = len(ds) // split_events
            if len(ds) % split_events:
                num_files += 1
            for ii in range(num_files):
                pp = path_out / f"{path_in.stem}_{ii+1:04d}.rtdc"
                pt = pp.with_suffix(".rtdc~")
                paths_gen.append(pp)
                paths_temp.append(pt)
                if verbose:
                    print(f"Generating {ii+1:d}/{num_files:d}: {pt}")
                ds.filter.manual[:] = False  # reset filter
                ds.filter.manual[ii*split_events:(ii+1)*split_events] = True
                common.skip_empty_image_events(
                    ds=ds,
                    initial=skip_initial_empty_image,
                    final=skip_final_empty_image)
                ds.apply_filter()
                ds.export.hdf5(
                    path=pt, features=ds.features_innate, filtered=True)

        if w:
            logs["dclab-split-warnings"] = common.assemble_warnings(w)
        sample_name = ds.config["experiment"]["sample"]

    # Add the logs and update sample name
    for ii, pt in enumerate(paths_temp):
        meta = {"experiment": {"sample": f"{sample_name} {ii+1}/{num_files}"}}
        with RTDCWriter(pt) as hw:
            for name in logs:
                hw.store_log(name, logs[name])
            hw.store_metadata(meta)

    for pt, pp in zip(paths_temp, paths_gen):
        pt.rename(pp)

    if ret_out_paths:
        return paths_gen


def split_parser():
    descr = "Split an RT-DC measurement file (.tdms or .rtdc) into multiple " \
            + "smaller .rtdc files."
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('path_in', metavar="PATH_IN", type=str,
                        help='Input path (.tdms or .rtdc file)')
    parser.add_argument('--path_out', metavar="PATH_OUT", type=str,
                        default="SAME",
                        help='Output directory (defaults to same directory)')
    parser.add_argument('--split-events', type=int, default=10000,
                        help='Maximum number of events in each output file')
    parser.add_argument('--include-empty-boundary-images',
                        dest='include_empty_boundary_images',
                        action='store_true',
                        help='In old versions of Shape-In, the first or last '
                             + 'images were sometimes not stored in the '
                             + 'resulting .avi file. In dclab, such images '
                             + 'are represented as zero-valued images. Set '
                             + 'this option, if you wish to include these '
                             + 'events with empty image data.')
    parser.set_defaults(include_empty_boundary_images=False)

    return parser



"""Test command-line interface"""
import hashlib
import sys
import tempfile
import time
import pathlib
import shutil

import dclab
from dclab import cli, new_dataset, rtdc_dataset

import h5py
import numpy as np
import pytest

from helper_methods import retrieve_data


def test_check_suffix_disabled_compress():
    path_in_o = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    path_in = path_in_o.with_suffix("")
    path_in_o.rename(path_in)
    assert path_in.suffix == ""
    with pytest.raises(ValueError, match="Unsupported file type"):
        cli.compress(path_in=path_in,
                     path_out=path_in.with_name("compressed.rtdc"))
    # but this should work:
    cli.compress(path_in=path_in,
                 path_out=path_in.with_name("compressed2.rtdc"),
                 check_suffix=False)


def test_check_suffix_disabled_condense():
    path_in_o = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    path_in = path_in_o.with_suffix("")
    path_in_o.rename(path_in)
    assert path_in.suffix == ""
    with pytest.raises(ValueError, match="Unsupported file type"):
        cli.condense(path_in=path_in,
                     path_out=path_in.with_name("condensed.rtdc"))
    # but this should work:
    cli.condense(path_in=path_in,
                 path_out=path_in.with_name("condensed2.rtdc"),
                 check_suffix=False)


def test_check_suffix_disabled_repack():
    path_in_o = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    path_in = path_in_o.with_suffix("")
    path_in_o.rename(path_in)
    assert path_in.suffix == ""
    with pytest.raises(ValueError, match="Unsupported file type"):
        cli.repack(path_in=path_in,
                   path_out=path_in.with_name("repacked.rtdc"))
    # but this should work:
    cli.repack(path_in=path_in,
               path_out=path_in.with_name("repacked2.rtdc"),
               check_suffix=False)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_compress():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("compressed.rtdc")

    cli.compress(path_out=path_out, path_in=path_in)
    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert "dclab-compress" in dsj.logs
        assert len(dsj)
        assert len(dsj) == len(ds0)
        for feat in ds0.features:
            if feat in ["contour", "image", "mask"]:
                for ii in range(len(dsj)):
                    assert np.all(dsj[feat][ii] == ds0[feat][ii]), feat
            else:
                assert np.all(dsj[feat] == ds0[feat]), feat


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_compress_already_compressed():
    """By default, an already compressed dataset should not be compressed"""
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out1 = path_in.with_name("compressed_1.rtdc")
    path_out2 = path_in.with_name("compressed_2.rtdc")
    path_out3 = path_in.with_name("compressed_copy_of_1.rtdc")
    # this is straight-forward
    cli.compress(path_out=path_out1, path_in=path_in)
    # just for the sake of comparison
    time.sleep(1)  # we need different time stamps in path_out2
    cli.compress(path_out=path_out2, path_in=path_in)
    # this is not trivial
    cli.compress(path_out=path_out3, path_in=path_out1)

    # the first two files should not be the same (dates are written, etc)
    h1 = hashlib.md5(path_out1.read_bytes()).hexdigest()
    h2 = hashlib.md5(path_out2.read_bytes()).hexdigest()
    h3 = hashlib.md5(path_out3.read_bytes()).hexdigest()
    assert h1 != h2
    assert h1 == h3


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_compress_already_compressed_force():
    """An extension of the above test to make sure "force" works"""
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out1 = path_in.with_name("compressed_1.rtdc")
    path_out2 = path_in.with_name("compressed_not_a_copy_of_1.rtdc")
    # this is straight-forward
    cli.compress(path_out=path_out1, path_in=path_in)
    # just for the sake of comparison
    cli.compress(path_out=path_out2, path_in=path_out1, force=True)

    # the first two files should not be the same (dates are written, etc)
    h1 = hashlib.md5(path_out1.read_bytes()).hexdigest()
    h2 = hashlib.md5(path_out2.read_bytes()).hexdigest()
    assert h1 != h2


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_compress_with_online_polygon_filters():
    """Shape-In 2.3 supports online polygon filters"""
    path = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # add an artificial online polygon filter
    with h5py.File(path, "a") as h5:
        # set soft filter to True
        h5.attrs["online_filter:area_um,deform soft limit"] = True
        # set filter values
        pf_name = "online_filter:area_um,deform polygon points"
        area_um = h5["events"]["area_um"]
        deform = h5["events"]["deform"]
        pf_points = np.array([
            [np.mean(area_um) + np.std(area_um),
             np.mean(deform)],
            [np.mean(area_um) + np.std(area_um),
             np.mean(deform) + np.std(deform)],
            [np.mean(area_um),
             np.mean(deform) + np.std(deform)],
        ])
        h5.attrs[pf_name] = pf_points

    path_out = path.with_name("compressed.rtdc")
    cli.compress(path_out=path_out, path_in=path)

    with dclab.new_dataset(path_out) as ds:
        assert len(ds) == 8
        assert ds.config["online_filter"]["area_um,deform soft limit"]
        assert "area_um,deform polygon points" in ds.config["online_filter"]
        assert np.allclose(
            ds.config["online_filter"]["area_um,deform polygon points"],
            pf_points)


def test_compress_with_online_polygon_filters_real_data():
    """Shape-In 2.3 supports online polygon filters"""
    path = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")

    path_out = path.with_name("compressed.rtdc")
    cli.compress(path_out=path_out, path_in=path)

    with dclab.new_dataset(path_out) as ds:
        assert len(ds) == 1
        assert ds.config["online_filter"]["size_x,size_y soft limit"]
        assert "size_x,size_y polygon points" in ds.config["online_filter"]
        assert np.allclose(
            ds.config["online_filter"]["size_x,size_y polygon points"],
            [[0.1, 0.2],
             [0.1, 2.5],
             [3.3, 3.2],
             [5.2, 0.9]]
        )


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_condense():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("condensed.rtdc")

    cli.condense(path_out=path_out, path_in=path_in)
    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert "dclab-condense" in dsj.logs
        assert len(dsj)
        assert len(dsj) == len(ds0)
        for feat in dsj.features:
            assert np.all(dsj[feat] == ds0[feat])


def test_method_available():
    # DCOR depotize needs this
    assert hasattr(cli, "get_job_info")
    assert hasattr(cli, "get_command_log")


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_repack_basic():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("repacked.rtdc")

    cli.repack(path_out=path_out, path_in=path_in)

    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert len(dsj)
        assert len(dsj) == len(ds0)
        for feat in ds0.features_innate:
            if feat in ds0.features_scalar:
                assert np.all(dsj[feat] == ds0[feat]), feat
        for ii in range(len(ds0)):
            assert np.all(dsj["contour"][ii] == ds0["contour"][ii])
            assert np.all(dsj["image"][ii] == ds0["image"][ii])
            assert np.all(dsj["mask"][ii] == ds0["mask"][ii])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_repack_remove_secrets():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("repacked.rtdc")

    with h5py.File(path_in, "a") as h5:
        h5.attrs["experiment:sample"] = "my dirty secret"

    with h5py.File(path_in, "a") as h5:
        h5.attrs["experiment:sample"] = "sunshine"

    # test whether the dirty secret is still there
    with open(str(path_in), "rb") as fd:
        data = fd.read()
        assert str(data).count("my dirty secret")

    # now repack
    cli.repack(path_out=path_out, path_in=path_in)

    # clean?
    with open(str(path_out), "rb") as fd:
        data = fd.read()
        assert not str(data).count("my dirty secret")


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_repack_strip_logs():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("repacked.rtdc")

    # write some logs
    with h5py.File(path_in, "a") as h5:
        hw = rtdc_dataset.RTDCWriter(h5)
        hw.store_log("test_log", ["peter", "hans"])

    cli.repack(path_out=path_out, path_in=path_in, strip_logs=True)

    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert ds0.logs
        assert not dsj.logs


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_repack_user_metadata():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    with h5py.File(path_in, "a") as h5:
        h5.attrs["user:peter"] = "hans"

    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("repacked.rtdc")

    cli.repack(path_out=path_out, path_in=path_in)

    with new_dataset(path_out) as ds:
        assert ds.config["user"]["peter"] == "hans"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_split():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    paths = cli.split(path_in=path_in, split_events=3, ret_out_paths=True)
    with new_dataset(path_in) as ds:
        ecount = 0
        for pp in paths:
            with new_dataset(pp) as di:
                for feat in ds.features_scalar:
                    if feat == "index":
                        continue
                    assert np.all(
                        ds[feat][ecount:ecount + len(di)] == di[feat]), feat
                ecount += len(di)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_split_traces():
    path_in = retrieve_data("fmt-hdf5_fl_2018.zip")
    paths = cli.split(path_in=path_in, split_events=3, ret_out_paths=True)
    with new_dataset(path_in) as ds:
        ecount = 0
        for pp in paths:
            with new_dataset(pp) as di:
                for flkey in ds["trace"].keys():
                    trace1 = ds["trace"][flkey][ecount:ecount + len(di)]
                    trace2 = di["trace"][flkey][:]
                    assert len(trace1) == len(trace2)
                    assert np.all(trace1 == trace2), flkey
                ecount += len(di)


def test_tdms2rtdc():
    pytest.importorskip("nptdms")
    path_in = retrieve_data("fmt-tdms_shapein-2.0.1-no-image_2017.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("out.rtdc")

    cli.tdms2rtdc(path_tdms=path_in,
                  path_rtdc=path_out,
                  compute_features=False)

    with new_dataset(path_out) as ds2, new_dataset(path_in) as ds1:
        assert len(ds2)
        assert set(ds1.features) == set(ds2.features)
        # not all features are computed
        assert set(ds2._events.keys()) < set(ds1.features)
        for feat in ds1:
            assert np.all(ds1[feat] == ds2[feat])


def test_tdms2rtdc_bulk():
    pytest.importorskip("nptdms")
    path_data = retrieve_data("fmt-tdms_shapein-2.0.1-no-image_2017.zip")
    path_wd = pathlib.Path(
        tempfile.mkdtemp(prefix="tdms2rtdc_bulk_")).resolve()
    path_in = path_wd / "input"
    path_in.mkdir()
    shutil.copytree(path_data.parent, path_in / "data_1")
    shutil.copytree(path_data.parent, path_in / "data_2")
    shutil.copytree(path_data.parent, path_in / "data_3")
    (path_in / "data_nested").mkdir()
    shutil.copytree(path_data.parent, path_in / "data_nested" / "data_4")
    # same directory (will be cleaned up with path_in)
    path_out = path_wd / "output"
    path_out.mkdir()

    cli.tdms2rtdc(path_tdms=path_in,
                  path_rtdc=path_out,
                  compute_features=False)

    for pp in [path_out / "data_1" / "M1_data.rtdc",
               path_out / "data_2" / "M1_data.rtdc",
               path_out / "data_3" / "M1_data.rtdc",
               path_out / "data_nested" / "data_4" / "M1_data.rtdc"]:
        assert pp.exists()

        with new_dataset(pp) as ds2, new_dataset(path_data) as ds1:
            assert len(ds2)
            assert set(ds1.features) == set(ds2.features)
            # not all features are computed
            assert set(ds2._events.keys()) < set(ds1.features)
            for feat in ds1:
                assert np.all(ds1[feat] == ds2[feat])


def test_tdms2rtdc_features():
    pytest.importorskip("nptdms")
    path_in = retrieve_data("fmt-tdms_shapein-2.0.1-no-image_2017.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("out.rtdc")

    cli.tdms2rtdc(path_tdms=path_in,
                  path_rtdc=path_out,
                  compute_features=True)

    with new_dataset(path_out) as ds2, new_dataset(path_in) as ds1:
        assert len(ds2)
        assert set(ds1.features) == set(ds2.features)
        # features were computed
        assert set(ds2._events.keys()) == set(ds1.features)


def test_tdms2rtdc_remove_nan_image():
    pytest.importorskip("nptdms")
    imageio = pytest.importorskip("imageio")
    path_in = retrieve_data("fmt-tdms_fl-image-bright_2017.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("out.rtdc")

    # generate fake video
    with new_dataset(path_in) as ds:
        video_length = len(ds) - 1
    vname = path_in.with_name("M4_0.040000ul_s_imaq.avi")
    # remove contour data (not necessary for this test)
    path_in.with_name("M4_0.040000ul_s_contours.txt").unlink()

    imgs = imageio.mimread(vname)
    with imageio.get_writer(vname) as writer:
        for ii in range(video_length):
            writer.append_data(imgs[ii % len(imgs)])

    # without removal
    cli.tdms2rtdc(path_tdms=path_in,
                  path_rtdc=path_out,
                  compute_features=False,
                  skip_initial_empty_image=False)

    with new_dataset(path_out) as ds2, new_dataset(path_in) as ds1:
        assert len(ds2) == len(ds1)
        assert np.all(ds2["image"][0] == 0)

    # with removal
    cli.tdms2rtdc(path_tdms=path_in,
                  path_rtdc=path_out,
                  compute_features=False,
                  skip_initial_empty_image=True)

    with new_dataset(path_out) as ds2, new_dataset(path_in) as ds1:
        assert len(ds2) == video_length
        assert not np.all(ds2["image"][0] == 0)
        assert ds2.config["experiment"]["event count"] == video_length


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_tdms2rtdc_update_roi_size():
    pytest.importorskip("nptdms")
    path_in = retrieve_data("fmt-tdms_fl-image_2016.zip")
    # set wrong roi sizes
    camin = path_in.with_name("M1_camera.ini")
    with camin.open("r") as fd:
        lines = fd.readlines()
    lines = lines[:-2]
    lines.append("width = 23\n")
    lines.append("height = 24\n")
    with camin.open("w") as fd:
        fd.writelines(lines)

    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("out.rtdc")

    cli.tdms2rtdc(path_tdms=path_in,
                  path_rtdc=path_out,
                  compute_features=False,
                  skip_initial_empty_image=True)

    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert ds0.config["imaging"]["roi size x"] == 23
        assert ds0.config["imaging"]["roi size y"] == 24
        assert dsj.config["imaging"]["roi size x"] == 256
        assert dsj.config["imaging"]["roi size y"] == 96
        wlog = "dclab-tdms2rtdc-warnings"
        assert "LimitingExportSizeWarning" in dsj.logs[wlog]


def test_tdms2rtdc_update_sample_per_events():
    pytest.importorskip("nptdms")
    path_in = retrieve_data("fmt-tdms_2fl-no-image_2017.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("out.rtdc")

    # set wrong samples per event
    with path_in.with_name("M1_para.ini").open("a") as fd:
        fd.write("Samples Per Event = 1234")

    with new_dataset(path_in) as ds:
        assert ds.config["fluorescence"]["samples per event"] == 1234

    cli.tdms2rtdc(path_tdms=path_in,
                  path_rtdc=path_out,
                  compute_features=True)

    with new_dataset(path_out) as ds2:
        assert ds2.config["fluorescence"]["samples per event"] == 566


@pytest.mark.parametrize("dataset,exit_status_expected", [
    ["fmt-hdf5_fl_2017.zip", 3],
    ["fmt-hdf5_fl_2018.zip", 1],
    ["fmt-hdf5_polygon_gate_2021.zip", 0],
])
def test_verify_dataset_exit_codes(dataset, exit_status_expected, monkeypatch):
    # get the exit status from the script
    def sys_exit(status):
        return status

    monkeypatch.setattr(sys, "exit", sys_exit)

    h5path = retrieve_data(dataset)
    exit_status = cli.verify_dataset(h5path)
    assert exit_status == exit_status_expected


def test_verify_dataset_exit_code_alert(monkeypatch):
    # get the exit status from the script
    def sys_exit(status):
        return status

    monkeypatch.setattr(sys, "exit", sys_exit)

    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")

    # Since 0.35.0 we really check the configuration key types.
    # Just make sure that they are properly set:
    with h5py.File(h5path, "a") as h5:
        for key in ["fluorescence:sample rate",
                    "imaging:roi position x",
                    "imaging:roi position y"]:
            h5.attrs[key] = int(h5.attrs[key])

    # provoke a warning
    with h5py.File(h5path, "r+") as h5:
        h5.attrs["experiment:unknown"] = ""
    exit_status = cli.verify_dataset(h5path)
    assert exit_status == 1  # unknown key leads to alert


def test_verify_dataset_exit_code_error(monkeypatch):
    # get the exit status from the script
    def sys_exit(status):
        return status

    monkeypatch.setattr(sys, "exit", sys_exit)

    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")

    # Since 0.35.0 we really check the configuration key types.
    # Just make sure that they are properly set:
    with h5py.File(h5path, "a") as h5:
        for key in ["fluorescence:sample rate",
                    "imaging:roi position x",
                    "imaging:roi position y"]:
            h5.attrs[key] = int(h5.attrs[key])

    # provoke a warning
    with h5py.File(h5path, "r+") as h5:
        h5.attrs["setup:channel width"] = "peter"
    exit_status = cli.verify_dataset(h5path)
    assert exit_status == 4  # Cannot convert string to float error


def test_verify_dataset_exit_code_user_ok(monkeypatch):
    # get the exit status from the script
    def sys_exit(status):
        return status

    monkeypatch.setattr(sys, "exit", sys_exit)

    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")

    # Since 0.35.0 we really check the configuration key types.
    # Just make sure that they are properly set:
    with h5py.File(h5path, "a") as h5:
        for key in ["fluorescence:sample rate",
                    "imaging:roi position x",
                    "imaging:roi position y"]:
            h5.attrs[key] = int(h5.attrs[key])

    # provoke a warning
    with h5py.File(h5path, "r+") as h5:
        h5.attrs["user:pan"] = 2
    exit_status = cli.verify_dataset(h5path)
    assert exit_status == 0


def test_verify_dataset_exit_code_violation_1(monkeypatch):
    # get the exit status from the script
    def sys_exit(status):
        return status

    monkeypatch.setattr(sys, "exit", sys_exit)

    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")

    # Since 0.35.0 we really check the configuration key types.
    # Just make sure that they are properly set:
    with h5py.File(h5path, "a") as h5:
        for key in ["fluorescence:sample rate",
                    "imaging:roi position x",
                    "imaging:roi position y"]:
            h5.attrs[key] = int(h5.attrs[key])

    # provoke a warning
    with h5py.File(h5path, "r+") as h5:
        h5.attrs["setup:flow rate"] = 0
    exit_status = cli.verify_dataset(h5path)
    assert exit_status == 3  # zero flow rate and warnings about mismatch


def test_verify_dataset_exit_code_violation_2(monkeypatch):
    # get the exit status from the script
    def sys_exit(status):
        return status

    monkeypatch.setattr(sys, "exit", sys_exit)

    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")

    # Since 0.35.0 we really check the configuration key types.
    # Just make sure that they are properly set:
    with h5py.File(h5path, "a") as h5:
        for key in ["fluorescence:sample rate",
                    "imaging:roi position x",
                    "imaging:roi position y"]:
            h5.attrs[key] = int(h5.attrs[key])

    # provoke a warning
    with h5py.File(h5path, "r+") as h5:
        h5.attrs["setup:flow rate"] = 0
        h5.attrs["setup:flow rate sample"] = 0
        h5.attrs["setup:flow rate sheath"] = 0
    exit_status = cli.verify_dataset(h5path)
    assert exit_status == 2  # zero flow rate



"""Class for on-the-fly conversion of contours to masks"""
import functools
import numbers

import numpy as np
import scipy.ndimage as ndi


class MaskColumn(object):
    def __init__(self, rtdc_dataset):
        """Computes mask from contour data"""
        self.contour = rtdc_dataset["contour"]
        self.image = rtdc_dataset["image"]
        self.identifier = self.contour.identifier
        self.config = rtdc_dataset.config

    def __getitem__(self, idx):
        if not isinstance(idx, numbers.Integral):
            raise NotImplementedError(
                "The RTDC_TDMS data handler does not support indexing with "
                "anything else than scalar integers. Please convert your data "
                "to the .rtdc file format!")

        mask = np.zeros(self._img_shape, dtype=bool)
        conti = self.contour[idx]
        mask[conti[:, 1], conti[:, 0]] = True
        ndi.morphology.binary_fill_holes(mask, output=mask)
        return mask

    def __len__(self):
        if self._img_shape != (0, 0):
            lc = len(self.contour)
        else:
            lc = 0
        return lc

    @property
    @functools.lru_cache()
    def _img_shape(self):
        """Shape of one event image"""
        cfgim = self.config["imaging"]
        if self.image:
            # get shape from image column
            event_image_shape = self.image.shape[1:]
        elif "roi size x" in cfgim and "roi size y" in cfgim:
            # get shape from config (this is less reliable than getting
            # the shape from the image; there were measurements with
            # wrong config keys)
            event_image_shape = (cfgim["roi size y"], cfgim["roi size x"])
        else:
            # no shape available
            event_image_shape = (0, 0)
        return event_image_shape

    @property
    @functools.lru_cache()
    def shape(self):
        return len(self), self._img_shape[0], self._img_shape[1]



import pathlib
import tempfile
import traceback

import h5py
import numpy as np
import pytest
from scipy.ndimage.filters import gaussian_filter

import dclab
from dclab.rtdc_dataset.feat_anc_plugin.plugin_feature import (
    PlugInFeature, import_plugin_feature_script,
    remove_plugin_feature, remove_all_plugin_features,
    PluginImportError)
from dclab.rtdc_dataset.feat_anc_core.ancillary_feature import (
    BadFeatureSizeWarning)
from dclab.rtdc_dataset.feat_temp import deregister_all
from dclab.rtdc_dataset.fmt_hierarchy import ChildNDArray

from helper_methods import retrieve_data

data_dir = pathlib.Path(__file__).parent / "data"


@pytest.fixture(autouse=True)
def cleanup_plugin_features():
    """Fixture used to cleanup plugin feature tests"""
    # code run before the test
    pass
    # then the test is run
    yield
    # code run after the test
    # remove our test plugin examples
    remove_all_plugin_features()
    deregister_all()


def compute_single_plugin_feature(rtdc_ds):
    """Basic plugin method"""
    circ_per_area = rtdc_ds["circ"] / rtdc_ds["area_um"]
    return circ_per_area


def compute_multiple_plugin_features(rtdc_ds):
    """Basic plugin method with dictionary returned"""
    circ_per_area = rtdc_ds["circ"] / rtdc_ds["area_um"]
    circ_times_area = rtdc_ds["circ"] * rtdc_ds["area_um"]
    return {"circ_per_area": circ_per_area, "circ_times_area": circ_times_area}


def compute_non_scalar_plugin_feature(rtdc_ds):
    """Basic non-scalar plugin method"""
    image_gauss_filter = gaussian_filter(rtdc_ds["image"], sigma=(0, 1, 1))
    return {"image_gauss_filter": image_gauss_filter}


def example_plugin_info_single_feature():
    """plugin info for a single feature"""
    info = {
        "method": compute_single_plugin_feature,
        "description": "This plugin will compute a feature",
        "long description": "Even longer description that "
                            "can span multiple lines",
        "feature names": ["circ_per_area"],
        "feature labels": ["Circularity per Area"],
        "features required": ["circ", "area_um"],
        "config required": [],
        "method check required": lambda x: True,
        "scalar feature": [True],
        "version": "0.1.0",
    }
    return info


def example_plugin_info_multiple_feature():
    """plugin info for multiple features"""
    info = {
        "method": compute_multiple_plugin_features,
        "description": "This plugin will compute some features",
        "long description": "Even longer description that "
                            "can span multiple lines",
        "feature names": ["circ_per_area", "circ_times_area"],
        "feature labels": ["Circularity per Area", "Circularity times Area"],
        "features required": ["circ", "area_um"],
        "config required": [],
        "method check required": lambda x: True,
        "scalar feature": [True, True],
        "version": "0.1.0",
    }
    return info


def example_plugin_info_non_scalar_feature():
    """plugin info for non-scalar feature"""
    info = {
        "method": compute_non_scalar_plugin_feature,
        "description": "This plugin will compute a non-scalar feature",
        "long description": "This non-scalar feature is a Gaussian filter of "
                            "the image",
        "feature names": ["image_gauss_filter"],
        "feature labels": ["Gaussian Filtered Image"],
        "features required": ["image"],
        "config required": [],
        "method check required": lambda x: True,
        "scalar feature": [False],
        "version": "0.1.0",
    }
    return info


def compute_with_user_section(rtdc_ds):
    """setup a plugin method that uses user config section

    The "user:n_constrictions" metadata must be set
    """
    nc = rtdc_ds.config["user"]["n_constrictions"]
    assert isinstance(nc, int), (
        '"n_constrictions" should be an integer value.')
    area_of_region = rtdc_ds["area_um"] * nc
    return {"area_of_region": area_of_region}


def test_pf_attribute_ancill_info():
    """Check the plugin feature attribute input to AncillaryFeature"""
    info = example_plugin_info_single_feature()
    pf = PlugInFeature("circ_per_area", info)
    assert pf.plugin_feature_info["feature name"] == "circ_per_area"
    assert pf.plugin_feature_info["method"] is compute_single_plugin_feature
    assert pf.plugin_feature_info["config required"] == []
    assert pf.plugin_feature_info["features required"] == ["circ", "area_um"]


def test_pf_attribute_plugin_feature_info():
    """Check the plugin feature info attribute"""
    info = example_plugin_info_single_feature()
    # comparing lambda functions fails due to differing memory locations
    info.pop("method check required")
    pf = PlugInFeature("circ_per_area", info)
    pf.plugin_feature_info.pop("method check required")
    plugin_feature_info = {
        "method": compute_single_plugin_feature,
        "description": "This plugin will compute a feature",
        "long description": "Even longer description that "
                            "can span multiple lines",
        "feature name": "circ_per_area",
        "feature label": "Circularity per Area",
        "feature shape": (1,),
        "features required": ["circ", "area_um"],
        "config required": [],
        "scalar feature": True,
        "version": "0.1.0",
        "plugin path": None,
        "identifier": "3a3e72c4cb015424ebbe6d4af63f2170",
    }
    assert pf.plugin_feature_info == plugin_feature_info


def test_pf_attributes():
    """Check the plugin feature attributes"""
    plugin_path = data_dir / "feat_anc_plugin_creative.py"
    plugin_list = dclab.load_plugin_feature(plugin_path)
    pf1, pf2 = plugin_list
    plugin_file_info = import_plugin_feature_script(plugin_path)

    assert pf1.feature_name == pf1.feature_name == \
        plugin_file_info["feature names"][0]
    assert pf2.feature_name == pf2.feature_name == \
        plugin_file_info["feature names"][1]

    assert plugin_path.samefile(pf1.plugin_path)
    assert plugin_path.samefile(pf1.plugin_feature_info["plugin path"])

    assert plugin_path.samefile(pf2.plugin_path)
    assert plugin_path.samefile(pf2.plugin_feature_info["plugin path"])

    assert pf1._original_info == plugin_file_info
    assert pf2._original_info == plugin_file_info


def test_pf_attributes_af_inherited():
    """Check the plugin feature attributes inherited from AncillaryFeature"""
    plugin_path = data_dir / "feat_anc_plugin_creative.py"
    plugin_list = dclab.load_plugin_feature(plugin_path)
    pf, _ = plugin_list
    plugin_file_info = import_plugin_feature_script(plugin_path)

    assert pf.feature_name == plugin_file_info["feature names"][0]
    assert pf.method == plugin_file_info["method"]
    assert pf.req_config == plugin_file_info["config required"]
    assert pf.req_features == plugin_file_info["features required"]
    assert pf.req_func == plugin_file_info["method check required"]
    assert pf.priority == 0


def test_pf_bad_plugin_feature_name_list():
    """Basic test of a bad feature name for PlugInFeature"""
    info = example_plugin_info_single_feature()
    info["feature names"] = "Peter-Pan's Best Friend!"
    with pytest.raises(ValueError, match="must be a list, got"):
        PlugInFeature("Peter-Pan's Best Friend!", info)


def test_pf_bad_plugin_feature_name():
    """Basic test of a bad feature name for PlugInFeature"""
    info = example_plugin_info_single_feature()
    info["feature names"] = ["Peter-Pan's Best Friend!"]
    with pytest.raises(ValueError, match="only contain lower-case characters"):
        PlugInFeature("Peter-Pan's Best Friend!", info)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_exists_in_hierarchy():
    """Test that RTDCHierarchy works with PlugInFeature"""
    info = example_plugin_info_single_feature()
    pf = PlugInFeature("circ_per_area", info)
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        assert pf.feature_name in ds
        assert dclab.dfn.feature_exists(pf.feature_name)
        child = dclab.new_dataset(ds)
        assert pf.feature_name in child


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_export_and_load():
    """Check that exported and loaded hdf5 file will keep a plugin feature"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    # initialize PlugInFeature instance
    info = example_plugin_info_single_feature()
    pf = PlugInFeature("circ_per_area", info)

    with dclab.new_dataset(h5path) as ds:
        # extract the feature information from the dataset
        assert pf in PlugInFeature.features
        circ_per_area = ds[pf.feature_name]

        # export the data to a new file
        expath = h5path.with_name("exported.rtdc")
        ds.export.hdf5(expath, features=ds.features_innate + [pf.feature_name])

    # make sure that worked
    with h5py.File(expath, "r") as h5:
        assert pf.feature_name in h5["events"]
        assert np.allclose(h5["events"][pf.feature_name], circ_per_area)

    # now check again with dclab
    with dclab.new_dataset(expath) as ds2:
        assert pf in PlugInFeature.features
        assert pf.feature_name in ds2
        assert pf.feature_name in ds2.features_innate
        assert np.allclose(ds2[pf.feature_name], circ_per_area)

        # and a control check
        remove_plugin_feature(pf)
        assert pf.feature_name not in ds2


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_export_non_scalar():
    h5path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    # initialize PlugInFeature instance
    info = example_plugin_info_non_scalar_feature()
    pf = PlugInFeature("image_gauss_filter", info)

    with dclab.new_dataset(h5path) as ds:
        # extract the feature information from the dataset
        assert pf in PlugInFeature.features
        image_gauss_filter = ds[pf.feature_name]

        # export the data to a new file
        expath = h5path.with_name("exported.rtdc")
        ds.export.hdf5(expath, features=[pf.feature_name])

    # make sure that worked
    with h5py.File(expath, "r") as h5:
        assert pf.feature_name in h5["events"]
        assert np.allclose(h5["events"][pf.feature_name], image_gauss_filter)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_export_non_scalar_filtered_issue_166():
    h5path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    # initialize PlugInFeature instance
    info = example_plugin_info_non_scalar_feature()
    info["feature shapes"] = [(80, 250)]
    pf = PlugInFeature("image_gauss_filter", info)

    with dclab.new_dataset(h5path) as ds:
        # extract the feature information from the dataset
        assert pf in PlugInFeature.features
        image_gauss_filter = ds[pf.feature_name]
        assert np.all(ds.filter.manual)  # just checking the default
        ds.filter.manual[0] = False
        ds.filter.manual[2] = False
        image_gauss_filter_filtered = np.array(
            image_gauss_filter[ds.filter.manual],
            copy=True)
        ds.apply_filter()

        # export the data to a new file
        expath = h5path.with_name("exported.rtdc")
        ds.export.hdf5(expath, features=[pf.feature_name], filtered=True)

    # make sure that worked
    with h5py.File(expath, "r") as h5:
        assert pf.feature_name in h5["events"]
        assert np.allclose(h5["events"][pf.feature_name],
                           image_gauss_filter_filtered)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_export_non_scalar_filtered_from_file_issue_166():
    h5path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    expath = h5path.with_name("exported.rtdc")
    # initialize PlugInFeature instance
    info = example_plugin_info_non_scalar_feature()
    info["feature shapes"] = [(80, 250)]
    pf = PlugInFeature("image_gauss_filter", info)

    # write the plugin feature data to an HDF5 file
    with dclab.new_dataset(h5path) as ds:
        # extract the feature information from the dataset
        ds.export.hdf5(expath, features=[pf.feature_name])

    # remove all plugin features and work with temporary feature
    remove_all_plugin_features()

    # try to load the plugin feature data from that HDF5 file and
    # export it again (this time, the exporter has to get the data
    # from the H5File object).
    expath2 = h5path.with_name("exported2.rtdc")
    dclab.register_temporary_feature("image_gauss_filter", is_scalar=False)
    with dclab.new_dataset(expath) as ds2:
        ds2.export.hdf5(expath2,
                        features=["image_gauss_filter"],
                        filtered=True)

    # make sure that worked
    with h5py.File(expath2, "r") as h5:
        assert "image_gauss_filter" in h5["events"]


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_export_non_scalar_single_event():
    h5path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    # initialize PlugInFeature instance
    info = example_plugin_info_non_scalar_feature()
    info["feature shapes"] = [(80, 250)]
    pf = PlugInFeature("image_gauss_filter", info)

    with dclab.new_dataset(h5path) as ds:
        # extract the feature information from the dataset
        assert pf in PlugInFeature.features
        image_gauss_filter = ds[pf.feature_name]

        # export the data to a new file
        expath = h5path.with_name("exported.rtdc")
        ds.export.hdf5(expath, features=["image", pf.feature_name])
        # write another single event
        with dclab.RTDCWriter(expath) as hw:
            hw.store_feature(pf.feature_name, ds["image"][0])
            hw.store_feature("image", ds["image"][0])

    # make sure that worked
    with h5py.File(expath, "r") as h5:
        assert pf.feature_name in h5["events"]
        assert np.allclose(h5["events"][pf.feature_name][:-1],
                           image_gauss_filter)
        assert np.allclose(h5["events"][pf.feature_name][-1],
                           h5["events/image"][0])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_export_non_scalar_no_warning():
    h5path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    # initialize PlugInFeature instance
    info = example_plugin_info_non_scalar_feature()
    info["feature shapes"] = [(80, 250)]
    pf = PlugInFeature("image_gauss_filter", info)

    with dclab.new_dataset(h5path) as ds:
        # extract the feature information from the dataset
        assert pf in PlugInFeature.features
        image_gauss_filter = ds[pf.feature_name]

        # export the data to a new file
        expath = h5path.with_name("exported.rtdc")
        ds.export.hdf5(expath, features=[pf.feature_name])

    # make sure that worked
    with h5py.File(expath, "r") as h5:
        assert pf.feature_name in h5["events"]
        assert np.allclose(h5["events"][pf.feature_name], image_gauss_filter)


def test_pf_import_error_attribution_issue_160_1():
    """Test for PluginImportError (file not found)

    When plugin features cannot be imported/found, a PluginImportError is
    raised with a generic error message. However, if the plugin itself
    contains an import that is not available, the actual exception
    should be raised.
    """
    with pytest.raises(PluginImportError,
                       match="The plugin could be not be found"):
        import_plugin_feature_script("/does/not/exist/laksdlaisdlaism.py")


def test_pf_import_error_attribution_issue_160_2():
    """Test for bad import in plugin

    When plugin features cannot be imported/found, a PluginImportError is
    raised with a generic error message. However, if the plugin itself
    contains an import that is not available, the actual exception
    should be raised.
    """
    tdir = tempfile.mkdtemp(prefix="plugin_import_error_")
    # create a plugin with an import that is not possible
    plugin_path = pathlib.Path(tdir) / "bad_plugin.py"
    good_plugin = (data_dir / "feat_anc_plugin_creative.py").read_text()
    bad_plugin = "import unexistent_module_alsdalsdvnalksdnf\n" + good_plugin
    plugin_path.write_text(bad_plugin)
    with pytest.raises(PluginImportError,
                       match="bad_plugin.py could not be loaded"):
        import_plugin_feature_script(plugin_path)

    try:
        import_plugin_feature_script(plugin_path)
    except PluginImportError:
        tb = traceback.format_exc()
        assert "No module named 'unexistent_module_alsdalsdvnalksdnf'" in tb


def test_pf_feature_exists():
    """Basic check that the plugin feature name exists in definitions"""
    plugin_path = data_dir / "feat_anc_plugin_creative.py"
    plugin_list = dclab.load_plugin_feature(plugin_path)
    assert dclab.dfn.feature_exists(plugin_list[0].feature_name)
    assert dclab.dfn.feature_exists(plugin_list[1].feature_name)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_filtering_with_plugin_feature():
    """Filtering with plugin feature"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        info = example_plugin_info_single_feature()
        pf = PlugInFeature("circ_per_area", info)

        ds.config["filtering"][f"{pf.feature_name} min"] = 0.030
        ds.config["filtering"][f"{pf.feature_name} max"] = 0.031
        ds.apply_filter()
        assert np.sum(ds.filter.all) == 1
        assert ds.filter.all[4]


def test_pf_import_plugin_info():
    """Check the plugin test example info is a dict"""
    plugin_path = data_dir / "feat_anc_plugin_creative.py"
    info = import_plugin_feature_script(plugin_path)
    assert isinstance(info, dict)


def test_pf_import_plugin_info_bad_path():
    """Raise error when a bad pathname is given"""
    bad_plugin_path = "not/a/real/path/plugin.py"
    with pytest.raises(PluginImportError, match="could be not be found"):
        import_plugin_feature_script(bad_plugin_path)


def test_pf_incorrect_input_info():
    """Raise error when info is not a dictionary"""
    info = ["this", "is", "not", "a", "dict"]
    with pytest.raises(ValueError, match="must be a dict"):
        PlugInFeature("feature_1", info)


def test_pf_incorrect_input_feature_name():
    """Raise error when the feature_name doesn't match info feature name"""
    info = example_plugin_info_single_feature()
    # `feature_name` is "circ_per_area" in info
    with pytest.raises(ValueError, match="is not defined"):
        PlugInFeature("not_the_correct_name", info)


def test_pf_incorrect_input_method():
    """Raise error when method is not callable"""
    info = example_plugin_info_single_feature()
    # set `info["method"]` to something that isn't callable
    info["method"] = "this_is_a_string"
    with pytest.raises(ValueError, match="is not callable"):
        PlugInFeature("circ_per_area", info)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_inherited_non_scalar():
    """Non scalar inherited plugin features should be of class ChildNDArray"""
    info = example_plugin_info_non_scalar_feature()
    PlugInFeature("image_gauss_filter", info)
    with dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip")) as ds:
        ds.filter.manual[2] = False
        ch = dclab.new_dataset(ds)
        assert "image_gauss_filter" in ch
        assert isinstance(ch["image_gauss_filter"], ChildNDArray)
        assert len(ch["image_gauss_filter"].shape) == 3


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_inherited_scalar():
    """Scalar inherited PluginFeatures should be a 1D np.ndarray"""
    info = example_plugin_info_single_feature()
    PlugInFeature("circ_per_area", info)
    with dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip")) as ds:
        ds.filter.manual[2] = False
        ch = dclab.new_dataset(ds)
        assert "circ_per_area" in ch
        assert isinstance(ch["circ_per_area"], np.ndarray)
        assert ch["circ_per_area"].ndim == 1


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_initialize_plugin_after_loading():
    """plugin feature loads correctly after feature added to hdf5 file"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        circ_per_area = compute_single_plugin_feature(ds)
    with h5py.File(h5path, "a") as h5:
        h5["events"]["circ_per_area"] = circ_per_area
    with dclab.new_dataset(h5path) as ds:
        assert "circ_per_area" not in ds
        info = example_plugin_info_single_feature()
        PlugInFeature("circ_per_area", info)
        assert "circ_per_area" in ds
        assert "circ_per_area" in ds.features_innate


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_initialize_plugin_feature_single():
    """Check that single plugin feature exists independant of loaded dataset"""
    ds = dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    info = example_plugin_info_single_feature()
    PlugInFeature("circ_per_area", info)
    assert "circ_per_area" in ds

    circ_per_area = ds["circ_per_area"]
    assert np.allclose(circ_per_area, ds["circ"] / ds["area_um"])

    # check that PlugInFeature exists independent of loaded ds
    ds2 = dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    assert "circ_per_area" in ds2


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_initialize_plugin_feature_non_scalar():
    """Check that the non-scalar plugin feature works"""
    ds = dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    info = example_plugin_info_non_scalar_feature()
    PlugInFeature("image_gauss_filter", info)
    assert "image_gauss_filter" in ds

    image_gauss_filter = ds["image_gauss_filter"]
    assert np.allclose(image_gauss_filter,
                       gaussian_filter(ds["image"], sigma=(0, 1, 1)))


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_initialize_plugin_features_multiple():
    """Check multiple plugin features exist independant of loaded dataset"""
    ds = dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    assert "circ_per_area" not in ds.features_innate
    assert "circ_times_area" not in ds.features_innate
    info = example_plugin_info_multiple_feature()
    PlugInFeature("circ_per_area", info)
    PlugInFeature("circ_times_area", info)

    assert "circ_per_area" in ds
    assert "circ_times_area" in ds
    assert dclab.dfn.feature_exists("circ_per_area")
    assert dclab.dfn.feature_exists("circ_times_area")
    circ_per_area = ds["circ_per_area"]
    circ_times_area = ds["circ_times_area"]
    assert np.allclose(circ_per_area, ds["circ"] / ds["area_um"])
    assert np.allclose(circ_times_area, ds["circ"] * ds["area_um"])


def test_pf_input_no_feature_labels():
    """Check that feature labels are populated even if not given"""
    info = example_plugin_info_single_feature()
    info.pop("feature labels")
    feature_name = "circ_per_area"
    pf = PlugInFeature(feature_name, info)
    assert dclab.dfn.feature_exists(feature_name)
    label = dclab.dfn.get_feature_label(feature_name)
    assert label == "Plugin feature {}".format(feature_name)
    assert label == pf.plugin_feature_info["feature label"]


def test_pf_input_no_scalar_feature():
    """Check that scalar feature bools are populated even if not given"""
    info = example_plugin_info_single_feature()
    info.pop("scalar feature")
    pf = PlugInFeature("circ_per_area", info)
    assert pf.plugin_feature_info["scalar feature"]


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_load_plugin():
    """Basic check for loading a plugin feature via a script"""
    ds = dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    assert "circ_per_area" not in ds.features_innate
    assert "circ_times_area" not in ds.features_innate
    plugin_path = data_dir / "feat_anc_plugin_creative.py"
    plugin_list = dclab.load_plugin_feature(plugin_path)
    assert isinstance(plugin_list[0], PlugInFeature)
    assert isinstance(plugin_list[1], PlugInFeature)
    assert "circ_per_area" in ds
    assert "circ_times_area" in ds
    circ_per_area = ds["circ_per_area"]
    circ_times_area = ds["circ_times_area"]
    assert np.allclose(circ_per_area, ds["circ"] / ds["area_um"])
    assert np.allclose(circ_times_area, ds["circ"] * ds["area_um"])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_load_non_scalar_plugin_data():
    """Test loading non-scalar plugin data format"""
    ds = dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    info = example_plugin_info_non_scalar_feature()
    info["feature shapes"] = [(80, 250)]
    PlugInFeature("image_gauss_filter", info)

    # Accessing non-scalar plugin data without prior saving and storing data in
    # HDF5-format should return already computed data as np.ndarray
    assert isinstance(ds["image_gauss_filter"], np.ndarray)

    # Exporting the rtdc-file including the plugin-feature and then reloading
    # it should load the non-scalar plugin data as h5py.Dataset
    tdir = tempfile.mkdtemp()
    pdir = pathlib.Path(tdir)
    pfile = pdir / "tmp.rtdc"
    features = ds.features + ["image_gauss_filter"]
    ds.export.hdf5(pfile, features=features)
    ds2 = dclab.new_dataset(pfile)
    assert isinstance(ds2["image_gauss_filter"], h5py.Dataset)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_load_scalar_plugin_data():
    """Test loading scalar plugin data return np.ndarray"""
    ds = dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    info = example_plugin_info_single_feature()
    PlugInFeature("circ_per_area", info)
    assert isinstance(ds["circ_per_area"], np.ndarray)

    # Exporting the rtdc-file including the plugin-feature and then reloading
    # it should still load scalar plugin data as np.ndarray
    tdir = tempfile.mkdtemp()
    pdir = pathlib.Path(tdir)
    pfile = pdir / "tmp.rtdc"
    features = ds.features + ["circ_per_area"]
    ds.export.hdf5(pfile, features=features)
    ds2 = dclab.new_dataset(pfile)
    assert isinstance(ds2["circ_per_area"], np.ndarray)


def test_pf_minimum_info_input():
    """Only method and feature names are required to create PlugInFeature"""
    info = {"method": compute_single_plugin_feature,
            "feature names": ["circ_per_area"]}
    pf = PlugInFeature("circ_per_area", info)

    # check that all other plugin_feature_info is populated
    assert "method" in pf.plugin_feature_info
    assert callable(pf.plugin_feature_info["method"])
    assert "description" in pf.plugin_feature_info
    assert "long description" in pf.plugin_feature_info
    assert "feature name" in pf.plugin_feature_info
    assert "feature label" in pf.plugin_feature_info
    assert "features required" in pf.plugin_feature_info
    assert "config required" in pf.plugin_feature_info
    assert "method check required" in pf.plugin_feature_info
    assert "scalar feature" in pf.plugin_feature_info
    assert "version" in pf.plugin_feature_info
    assert "plugin path" in pf.plugin_feature_info


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_remove_all_plugin_features():
    """Remove all plugin features at once"""
    ds = dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    assert "circ_per_area" not in ds.features_innate
    assert "circ_times_area" not in ds.features_innate
    plugin_path = data_dir / "feat_anc_plugin_creative.py"
    dclab.load_plugin_feature(plugin_path)
    assert "circ_per_area" in ds
    assert "circ_times_area" in ds
    assert dclab.dfn.feature_exists("circ_per_area")
    assert dclab.dfn.feature_exists("circ_times_area")

    remove_all_plugin_features()

    assert "circ_per_area" not in ds
    assert "circ_times_area" not in ds
    assert not dclab.dfn.feature_exists("circ_per_area")
    assert not dclab.dfn.feature_exists("circ_times_area")


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_remove_plugin_feature():
    """Remove individual plugin features"""
    ds = dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    assert "circ_per_area" not in ds
    assert "circ_times_area" not in ds

    plugin_path = data_dir / "feat_anc_plugin_creative.py"
    plugin_list = dclab.load_plugin_feature(plugin_path)
    assert len(plugin_list) == 2
    assert "circ_per_area" in ds
    assert "circ_per_area" not in ds.features_innate
    assert "circ_times_area" in ds
    assert "circ_times_area" not in ds.features_innate
    assert dclab.dfn.feature_exists("circ_per_area")
    assert dclab.dfn.feature_exists("circ_times_area")

    remove_plugin_feature(plugin_list[0])
    remove_plugin_feature(plugin_list[1])

    assert "circ_per_area" not in ds
    assert "circ_times_area" not in ds
    assert not dclab.dfn.feature_exists("circ_per_area")
    assert not dclab.dfn.feature_exists("circ_times_area")

    with pytest.raises(TypeError,
                       match="hould be an instance of PlugInFeature"):
        not_a_plugin_instance = [4, 6, 5]
        remove_plugin_feature(not_a_plugin_instance)


def test_pf_try_existing_feature_fails():
    """An existing feature name is not allowed"""
    info = example_plugin_info_single_feature()
    info["feature names"] = ["deform"]
    with pytest.raises(ValueError, match="Feature 'deform' already exists"):
        PlugInFeature("deform", info)


def test_pf_with_empty_feature_label_string():
    """An empty string is replaced with a real feature label

    Show that an empty `feature_label` will still give a descriptive
    feature label. See `dclab.dfn._add_feature_to_definitions` for details.
    """
    info = example_plugin_info_single_feature()
    info["feature labels"] = [""]
    feature_name = "circ_per_area"
    PlugInFeature(feature_name, info)
    assert dclab.dfn.feature_exists("circ_per_area")
    label = dclab.dfn.get_feature_label("circ_per_area")
    assert label != ""
    assert label == "Plugin feature {}".format(feature_name)


def test_pf_with_feature_label():
    """Check that a plugin feature label is added to definitions"""
    info = example_plugin_info_single_feature()
    info["feature labels"] = ["Circ / Area [1/µm²]"]
    feature_name = "circ_per_area"
    PlugInFeature(feature_name, info)
    assert dclab.dfn.feature_exists("circ_per_area")
    label = dclab.dfn.get_feature_label("circ_per_area")
    assert label == "Circ / Area [1/µm²]"


def test_pf_with_no_feature_label():
    """A feature label of None is replaced with a real feature label

    Show that `feature_label=None` will still give a descriptive
    feature label. See `dclab.dfn._add_feature_to_definitions` for details.
    """
    info = example_plugin_info_single_feature()
    info["feature labels"] = [None]
    feature_name = "circ_per_area"
    PlugInFeature(feature_name, info)
    assert dclab.dfn.feature_exists("circ_per_area")
    label = dclab.dfn.get_feature_label("circ_per_area")
    assert label is not None
    assert label == "Plugin feature {}".format(feature_name)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_with_user_config_section():
    """Use a plugin feature with the user defined config section"""
    info = {"method": compute_with_user_section,
            "feature names": ["area_of_region"],
            "config required": [["user", ["n_constrictions"]]]}
    PlugInFeature("area_of_region", info)

    ds = dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    assert "area_of_region" not in ds, "not available b/c missing metadata"
    # add some metadata to the user config section
    metadata = {"channel": True,
                "n_constrictions": 3}
    ds.config["user"].update(metadata)
    assert ds.config["user"] == metadata
    assert "area_of_region" in ds, "available b/c metadata is set"

    area_of_region1 = ds["area_of_region"]
    area_of_region1_calc = (ds["area_um"] *
                            ds.config["user"]["n_constrictions"])
    assert np.allclose(area_of_region1, area_of_region1_calc)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_with_user_config_section_fails():
    """Use a plugin feature with the user defined config section"""
    info = {"method": compute_with_user_section,
            "feature names": ["area_of_region"],
            "config required": [["user", ["n_constrictions"]]]}
    PlugInFeature("area_of_region", info)

    ds = dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    # show that the plugin feature is not available before setting the
    # user metadata
    ds.config["user"].clear()
    with pytest.raises(KeyError,
                       match=r"Feature \'area_of_region\' does not exist"):
        ds["area_of_region"]
    # show that the plugin fails when the user metadata type is wrong
    ds.config["user"]["n_constrictions"] = 4.99
    with pytest.raises(AssertionError, match="should be an integer value"):
        ds["area_of_region"]


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_wrong_data_shape_1():
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        info = example_plugin_info_single_feature()
        info["scalar feature"] = [False]
        pf = PlugInFeature("circ_per_area", info)
        with pytest.raises(ValueError, match="is not a scalar feature"):
            ds[pf.feature_name]


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_wrong_data_shape_2():
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        info = example_plugin_info_single_feature()
        info["scalar feature"] = [True]
        info["method"] = lambda x: np.arange(len(ds) * 2).reshape(-1, 2)
        pf = PlugInFeature("circ_per_area", info)
        with pytest.raises(ValueError, match="is a scalar feature"):
            ds[pf.feature_name]


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_wrong_length_1():
    """plugin feature should have same length"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        info = example_plugin_info_single_feature()
        info["method"] = lambda x: np.arange(len(ds) // 2)
        pf = PlugInFeature("circ_per_area", info)
        with pytest.warns(BadFeatureSizeWarning,
                          match="to match event number"):
            ds[pf.feature_name]


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_pf_wrong_length_2():
    """plugin feature should have same length"""
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        info = example_plugin_info_single_feature()
        info["method"] = lambda x: np.arange(len(ds) * 2)
        pf = PlugInFeature("circ_per_area", info)
        with pytest.warns(BadFeatureSizeWarning,
                          match="to match event number"):
            ds[pf.feature_name]


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()
            remove_all_plugin_features()
